# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

from __future__ import absolute_import, division, print_function

import json
import os

from six import iterkeys
from six.moves import configparser
from six.moves import BaseHTTPServer as httpserver  # NOQA


def _load_media_types():
    current_dir = os.path.dirname(os.path.realpath(__file__))
    data_dir = os.path.join(current_dir, "data")
    media_types_file = os.path.join(data_dir, "supported_mime_types.json")

    with open(media_types_file, "r") as f:
        return json.load(f)

HTTP_METHODS = [
    "get", "post", "put", "delete", "patch", "head",
    "options", "trace", "connect"
]
HTTP_OPTIONAL = [m + "?" for m in HTTP_METHODS].extend(HTTP_METHODS)

RAML_VERSIONS = ["0.8"]
PROTOCOLS = ["HTTP", "HTTPS"]
MEDIA_TYPES = _load_media_types()
AUTH_SCHEMES = [
    "oauth_1_0", "oauth_2_0",
    "basic", "basic_auth", "basicAuth", "basicAuthentication",
    "http_basic", "basic_authentication",
    "digest", "digest_auth", "digestAuth", "digestAuthentication",
    "digest_authentication", "http_digest"
]
HTTP_RESP_CODES = list(iterkeys(httpserver.BaseHTTPRequestHandler.responses))
PRIM_TYPES = ["string", "integer", "number", "boolean", "date", "file"]

CONFIG_VARS = [
    "auth_schemes", "resp_codes", "media_types", "protocols", "http_methods",
    "prim_types", "raml_versions"
]


def _clean(a_list):
    return sorted(list(set(a_list)))


def add_custom_config(user_config, parser_config):
    """Add user-defined config"""
    pc = parser_config

    if user_config.has_section("custom"):
        items = user_config.items("custom")
        for i in items:
            if i[0] not in CONFIG_VARS:
                continue
            conf = user_config.get("custom", i[0]).strip().split(",")
            conf = [c.strip() for c in conf]
            pc[i[0]] = pc[i[0]] + conf
    pc["resp_codes"] = _clean([int(r) for r in pc["resp_codes"]])
    pc["validate"] = user_config.has_option("main", "validate") and \
        user_config.get("main", "validate")
    pc["production"] = user_config.has_option("main", "production") and \
        user_config.get("main", "production")

    return pc


def setup_config(config_file=None):
    """
    Setup configuration for valid RAML parsing.

    :param file config_file: ``.ini`` file to be parsed (optional)
    :returns: ConfigParser instance
    """
    parser_config = {
        "auth_schemes": AUTH_SCHEMES,
        "resp_codes": HTTP_RESP_CODES,
        "media_types": MEDIA_TYPES,
        "protocols": PROTOCOLS,
        "http_methods": HTTP_METHODS,
        "raml_versions": RAML_VERSIONS,
        "prim_types": PRIM_TYPES,
        "validate": True,
    }

    if config_file:
        if not os.path.isfile(config_file):
            msg = "No such file or directory: '{0}'".format(config_file)
            raise IOError(msg)
        user_config = configparser.RawConfigParser()
        user_config.read(config_file)
        parser_config = add_custom_config(user_config, parser_config)

    optional = [m + "?" for m in parser_config["http_methods"]]
    parser_config["http_optional"] = optional + parser_config["http_methods"]
    return parser_config



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

from __future__ import absolute_import, division, print_function

import attr
from six.moves import BaseHTTPServer as httpserver  # NOQA

from .parameters import Content
from .validate import *  # NOQA

HTTP_RESP_CODES = httpserver.BaseHTTPRequestHandler.responses.keys()
AVAILABLE_METHODS = [
    "get", "post", "put", "delete", "patch", "head", "options",
    "trace", "connect"
]

METHOD_PROPERTIES = [
    "headers", "body", "responses", "query_params", "form_params"
]


@attr.s
class RootNode(object):
    """
    API Root Node

    :param dict raw: dict of loaded RAML data
    :param str version: API version
    :param str base_uri: API's base URI
    :param list base_uri_params: parameters for base URI, or ``None``. \
        The order of ``base_uri_params`` will follow the order \
        defined in the :py:obj:`.RootNode.base_uri`.
    :param list uri_params: URI parameters that can apply to all resources, \
        or ``None``. The order of ``uri_params`` will follow the order \
        defined in the :py:obj:`.RootNode.base_uri`.
    :param list protocols: API-supported protocols, defaults to protocol \
        in ``base_uri``
    :param str title: API Title
    :param list docs: list of :py:class:`parameters.Documentation` objects, \
        or ``None``
    :param list schemas: list of dictionaries, or ``None``
    :param str media_type: default accepted request/response media type, \
        or ``None``
    :param list resource_types: list of :py:class:`ResourceTypeNode`, \
        or ``None``
    :param list traits: list of :py:class:`TraitNode`, or ``None``
    :param list security_schemes: list of \
        :py:class:`parameters.SecurityScheme` objects, or ``None``
    :param list resources: list of :py:class:`ResourceNode` objects, \
        or ``None``
    :param raml_obj: loaded :py:class:`raml.RAMLDict` object
    """
    raw              = attr.ib(repr=False)
    version          = attr.ib(repr=False, validator=root_version)
    base_uri         = attr.ib(repr=False, validator=root_base_uri)
    base_uri_params  = attr.ib(repr=False,
                               validator=root_base_uri_params)
    uri_params       = attr.ib(repr=False, validator=root_uri_params)
    protocols        = attr.ib(repr=False, validator=root_protocols)
    title            = attr.ib(validator=root_title)
    documentation    = attr.ib(repr=False, validator=root_docs)
    schemas          = attr.ib(repr=False, validator=root_schemas)
    media_type       = attr.ib(repr=False, validator=root_media_type)
    secured_by       = attr.ib(repr=False, validator=root_secured_by)
    resource_types   = attr.ib(repr=False, init=False)
    traits           = attr.ib(repr=False, init=False)
    security_schemes = attr.ib(repr=False, init=False)
    resources        = attr.ib(repr=False, init=False,
                               validator=root_resources)
    raml_obj         = attr.ib(repr=False)
    config           = attr.ib(repr=False,
                               validator=attr.validators.instance_of(dict))
    errors           = attr.ib(repr=False)


@attr.s
class BaseNode(object):
    """
    :param dict raw: The raw data parsed from the RAML file
    :param RootNode root: Back reference to the node's API root
    :param list headers: List of node's :py:class:`parameters.Header` \
        objects, or ``None``
    :param list body: List of node's :py:class:`parameters.Body` objects, \
        or ``None``
    :param list responses: List of node's :py:class:`parameters.Response`\
        objects, or ``None``
    :param list uri_params: List of node's :py:class:`parameters.URIParameter`\
        objects, or ``None``. The order of ``uri_params`` will follow the \
        order defined in the \
        :py:obj:`.ResourceNode.absolute_uri`.
    :param list base_uri_params: List of node's base \
        :py:obj:`parameters.URIParameter` objects, or ``None``. The order of \
        ``base_uri_params`` will follow the order defined in the \
        :py:attribute:`.ResourceNode.absolute_uri`.
    :param list query_params: List of node's \
        :py:obj:`parameters.QueryParameter` objects, or ``None``
    :param list form_params: List of node's \
        :py:class:`parameters.FormParameter` objects, or ``None``
    :param str media_type: Supported request MIME media type. Defaults to \
        :py:class:`RootNode`'s ``media_type``.
    :param str description: Description of node.
    :param list protocols: List of ``str`` 's of supported protocols. \
        Defaults to :py:class:`RootNode`'s ``protocols``.
    """
    root            = attr.ib(repr=False)
    headers         = attr.ib(repr=False)
    body            = attr.ib(repr=False)
    responses       = attr.ib(repr=False)
    uri_params      = attr.ib(repr=False)
    base_uri_params = attr.ib(repr=False)
    query_params    = attr.ib(repr=False)
    form_params     = attr.ib(repr=False)
    media_type      = attr.ib(repr=False)
    desc            = attr.ib(repr=False)
    protocols       = attr.ib(repr=False)
    errors          = attr.ib(repr=False)

    @property
    def description(self):
        return Content(self.desc)


@attr.s
class TraitNode(BaseNode):
    """
    RAML Trait object

    :param str name: Name of trait
    :param str usage: Usage of trait
    """
    name  = attr.ib()
    raw   = attr.ib(repr=False, validator=defined_trait)
    usage = attr.ib(repr=False)


@attr.s
class ResourceTypeNode(BaseNode):
    """
    RAML Resource Type object

    :param str name: Name of resource type
    :param str type: Name of inherited :py:class:`ResourceTypeNode` object,
        or ``None``.
    :param str method: Supported method. If ends in ``?``, parameters will \
        only be applied to assigned resource if resource implements this \
        method. Else, resource must implement the method.
    :param str usage: Usage of resource type, or ``None``
    :param bool optional: Inherited if resource defines method.
    :param list is\_: List of assigned trait names, or ``None``
    :param list traits: List of assigned :py:class:`TraitNode` objects, \
        or ``None``
    :param str secured_by: List of ``str`` s or ``dict`` s of assigned \
        security scheme, or ``None``. If a ``str``, the name of the security \
        scheme.  If a ``dict``, the key is the name of the scheme, the values \
        are the parameters assigned (e.g. relevant OAuth 2 scopes).
    :param list security_schemes: A list of assigned \
        :py:class:`parameters.SecurityScheme` objects, or ``None``.
    :param str display_name: User-friendly name of resource; \
        defaults to ``name``

    """
    name             = attr.ib()
    raw              = attr.ib(repr=False, validator=defined_resource_type)
    type             = attr.ib(repr=False, validator=assigned_res_type)
    method           = attr.ib(repr=False)
    usage            = attr.ib(repr=False)
    optional         = attr.ib(repr=False)
    is_              = attr.ib(repr=False, validator=assigned_traits)
    traits           = attr.ib(repr=False)
    secured_by       = attr.ib(repr=False)
    security_schemes = attr.ib(repr=False)
    display_name     = attr.ib(repr=False)


@attr.s
class ResourceNode(BaseNode):
    """
    Supported API-endpoint (“resource”)

    :param str name: Resource name
    :param ResourceNode parent: parent node object if any, or ``None``
    :param str method: HTTP method for resource, or ``None``
    :param str display_name: User-friendly name of resource; \
        defaults to ``name``
    :param str path: relative path of resource
    :param str absolute_uri: Absolute URI of resource: \
        :py:class:`RootNode`'s ``base_uri`` + ``path``
    :param list is\_: A list of ``str`` s or ``dict`` s of resource-assigned \
        traits, or ``None``
    :param list traits: A list of assigned :py:class:`TraitNode` objects, \
        or ``None``
    :param str type: The name of the assigned resource type, or ``None``
    :param ResourceTypeNode resource_type: The assigned \
        :py:class:`ResourceTypeNode` object
    :param list secured_by: A list of ``str`` s or ``dict`` s of resource-\
        assigned security schemes, or ``None``. If a ``str``, the name of the \
        security scheme.  If a ``dict``, the key is the name of the scheme, \
        the values are the parameters assigned (e.g. relevant OAuth 2 scopes).
    :param list security_schemes: A list of assigned \
        :py:class:`parameters.SecurityScheme` objects, or ``None``.
    """
    name             = attr.ib(repr=False)
    raw              = attr.ib(repr=False)
    parent           = attr.ib(repr=False)
    method           = attr.ib()
    display_name     = attr.ib(repr=False)
    path             = attr.ib()
    absolute_uri     = attr.ib(repr=False)
    is_              = attr.ib(repr=False, validator=assigned_traits)
    traits           = attr.ib(repr=False)
    type             = attr.ib(repr=False, validator=assigned_res_type)
    resource_type    = attr.ib(repr=False)
    secured_by       = attr.ib(repr=False)
    security_schemes = attr.ib(repr=False)

    def _inherit_type(self):
        for p in METHOD_PROPERTIES:
            inherited_prop = getattr(self.resource_type, p)
            resource_prop = getattr(self, p)
            if resource_prop and inherited_prop:
                for r in resource_prop:
                    r._inherit_type_properties(inherited_prop)



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB
import os

import pytest

from ramlfications.errors import LoadRAMLError
from ramlfications._helpers import _get_raml_object

from .base import EXAMPLES


@pytest.fixture(scope="session")
def raml_file():
    return os.path.join(EXAMPLES + "complete-valid-example.raml")


def test_raml_file_is_none():
    raml_file = None
    with pytest.raises(LoadRAMLError) as e:
        _get_raml_object(raml_file)
    msg = ("RAML file can not be 'None'.",)
    assert e.value.args == msg


def test_raml_file_object(raml_file):
    with open(raml_file) as f:
        raml_obj = _get_raml_object(f)
        assert raml_obj == f


def test_not_valid_raml_obj():
    invalid_obj = 1234
    with pytest.raises(LoadRAMLError) as e:
        _get_raml_object(invalid_obj)
    msg = (("Can not load object '{0}': Not a basestring type or "
           "file object".format(invalid_obj)),)
    assert e.value.args == msg



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

from __future__ import absolute_import, division, print_function

__author__ = "Lynn Root"
__version__ = "0.1.9"
__license__ = "Apache 2.0"

__email__ = "lynn@spotify.com"
__uri__ = "https://ramlfications.readthedocs.org"
__description__ = "A Python RAML parser"


from ramlfications.config import setup_config
from ramlfications.parser import parse_raml

from ramlfications._helpers import load_file, load_string


def load(raml_file):
    """
    Module helper function to load a RAML File using \
    :py:class:`.loader.RAMLLoader`.

    :param str raml_file: String path to RAML file
    :return: loaded RAML
    :rtype: dict
    :raises LoadRAMLError: If error occurred trying to load the RAML file
    """
    return load_file(raml_file)


def loads(raml_string):
    """
    Module helper function to load a RAML File using \
    :py:class:`.loader.RAMLLoader`.

    :param str raml_string: String of RAML data
    :return: loaded RAML
    :rtype: dict
    :raises LoadRAMLError: If error occurred trying to load the RAML file
    """
    return load_string(raml_string)


def parse(raml, config_file=None):
    """
    Module helper function to parse a RAML File.  First loads the RAML file
    with :py:class:`.loader.RAMLLoader` then parses with
    :py:func:`.parser.parse_raml` to return a :py:class:`.raml.RAMLRoot`
    object.

    :param raml: Either string path to the RAML file, a file object, or \
        a string representation of RAML.
    :param str config_file:  String path to desired config file, if any.
    :return: parsed API
    :rtype: RAMLRoot
    :raises LoadRAMLError: If error occurred trying to load the RAML file
        (see :py:class:`.loader.RAMLLoader`)
    :raises InvalidRootNodeError: API metadata is invalid according to RAML \
        `specification <http://raml.org/spec.html>`_.
    :raises InvalidResourceNodeError: API resource endpoint is invalid \
        according to RAML `specification <http://raml.org/spec.html>`_.
    :raises InvalidParameterError: Named parameter is invalid \
        according to RAML `specification <http://raml.org/spec.html>`_.
    """
    loader = load(raml)
    config = setup_config(config_file)
    return parse_raml(loader, config)


def validate(raml, config_file=None):
    """
    Module helper function to validate a RAML File.  First loads \
    the RAML file \
    with :py:class:`.loader.RAMLLoader` then validates with \
    :py:func:`.validate.validate_raml`.

    :param str raml: Either string path to the RAML file, a file object, \or
        a string representation of RAML.
    :param str config_file:  String path to desired config file, if any.
    :return: No return value if successful
    :raises LoadRAMLError: If error occurred trying to load the RAML file
        (see :py:class:`.loader.RAMLLoader`)
    :raises InvalidRootNodeError: API metadata is invalid according to RAML \
        `specification <http://raml.org/spec.html>`_.
    :raises InvalidResourceNodeError: API resource endpoint is invalid \
        according to RAML `specification <http://raml.org/spec.html>`_.
    :raises InvalidParameterError: Named parameter is invalid \
        according to RAML `specification <http://raml.org/spec.html>`_.
    :raises InvalidRAMLError: RAML file is invalid according to RAML \
        `specification <http://raml.org/spec.html>`_.
    """
    loader = load(raml)
    config = setup_config(config_file)
    config["validate"] = True
    parse_raml(loader, config)



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

from __future__ import absolute_import, division, print_function

import attr
import markdown2 as md

from .validate import *  # NOQA

HTTP_METHODS = [
    "get", "post", "put", "delete", "patch", "options",
    "head", "trace", "connect"
]

NAMED_PARAMS = [
    "desc", "type", "enum", "pattern", "minimum", "maximum", "example",
    "default", "required", "repeat", "display_name", "max_length",
    "min_length"
]


class Content(object):
    """
    Returns documentable content from the RAML file (e.g. Documentation
    content, description) in either raw or parsed form.

    :param str data: The raw/marked up content data.
    """
    def __init__(self, data):
        self.data = data

    @property
    def raw(self):
        """
        Return raw Markdown/plain text written in the RAML file
        """
        return self.data

    @property
    def html(self):
        """
        Returns parsed Markdown into HTML
        """
        return md.markdown(self.data)

    def __repr__(self):
        return self.raw


@attr.s
class BaseParameter(object):
    """
    Base parameter with properties defined by the RAML spec's \
    'Named Parameters' section.

    :param str name: The item name of parameter
    :param dict raw: All defined data of the item
    :param str description: Parameter description, or ``None``.
    :param str display_name: User-friendly name for display or documentation \
        purposes.  If ``displayName`` is not specified in RAML file, defaults \
        to ``name``.
    :param int min_length: Parameter value's minimum number of characters, or \
        ``None``. Only applies when primative ``type`` is ``string``.
    :param int max_length: Parameter value's maximum number of characters, or \
        ``None``. Only applies when primative ``type`` is ``string``.
    :param int minimum: Parameter's minimum value, or ``None``.  Only applies \
        when primative ``type`` is ``integer`` or ``number``.
    :param int maximum: Parmeter's maximum value, or ``None``.  Only applies \
        when primative ``type`` is ``integer`` or ``number``.
    :param example: Example value for property, or ``None``.  Type of \
        ``example`` will match that of ``type``.
    :param default: Default value for property, or ``None``. Type of \
        ``default`` will match that of ``type``.
    :param bool repeat: If parameter can be repeated.  Defaults to ``False``.
    :param str pattern: A regular expression that parameter of type \
        ``string`` must match, or ``None`` if not set.
    :param list enum: Array of valid parameter values, or ``None``.  Only \
        applies when primative ``type`` is ``string``.
    :param str type: Primative type of parameter. Defaults to ``string`` if \
        not set.
    """
    name         = attr.ib()
    raw          = attr.ib(repr=False,
                           validator=attr.validators.instance_of(dict))
    desc         = attr.ib(repr=False)
    display_name = attr.ib(repr=False)
    min_length   = attr.ib(repr=False, validator=string_type_parameter)
    max_length   = attr.ib(repr=False, validator=string_type_parameter)
    minimum      = attr.ib(repr=False, validator=integer_number_type_parameter)
    maximum      = attr.ib(repr=False, validator=integer_number_type_parameter)
    example      = attr.ib(repr=False)
    default      = attr.ib(repr=False)
    config       = attr.ib(repr=False,
                           validator=attr.validators.instance_of(dict))
    errors       = attr.ib(repr=False)
    repeat       = attr.ib(repr=False, default=False)
    pattern      = attr.ib(repr=False, default=None,
                           validator=string_type_parameter)
    enum         = attr.ib(repr=False, default=None,
                           validator=string_type_parameter)
    type         = attr.ib(repr=False, default="string")

    @property
    def description(self):
        if self.desc:
            return Content(self.desc)
        return None

    def _inherit_type_properties(self, inherited_param):
        for param in inherited_param:
            name = getattr(param, "name", getattr(param, "code", None))
            if name == self.name:
                for n in NAMED_PARAMS:
                    attr = getattr(self, n, None)
                    if attr is None:
                        attr = getattr(param, n, None)
                        setattr(self, n, attr)


@attr.s
class URIParameter(BaseParameter):
    """
    URI parameter with properties defined by the RAML specification's \
    "Named Parameters" section, e.g.: ``/foo/{id}`` where ``id`` is the \
    name of the URI parameter.
    """
    required = attr.ib(repr=False, default=True)


@attr.s
class QueryParameter(BaseParameter):
    """
    Query parameter with properties defined by the RAML specification's \
    "Named Parameters" section, e.g. ``/foo/bar?baz=123`` where ``baz`` \
    is the name of the query parameter.
    """
    required = attr.ib(repr=False, default=False)


@attr.s
class FormParameter(BaseParameter):
    """
    Form parameter with properties defined by the RAML specification's
    "Named Parameters" section. Example:

        ``curl -X POST https://api.com/foo/bar -d baz=123``

    where ``baz`` is the Form Parameter name.
    """
    required = attr.ib(repr=False, default=False)


class Documentation(object):
    """
    User documentation for the API.

    :param str title: Title of documentation.
    :param str content: Content of documentation.
    """
    def __init__(self, _title, _content):
        self._title = _title
        self._content = _content

    @property
    def title(self):
        return Content(self._title)

    @property
    def content(self):
        return Content(self._content)

    def __repr__(self):  # NOCOV
        return "Documentation(title='{0}')".format(self.title)


@attr.s
class Header(object):
    """
    Header with properties defined by the RAML spec's 'Named Parameters'
    section, e.g.:

        ``curl -H 'X-Some-Header: foobar' ...``

    where ``X-Some-Header`` is the Header name.


    :param str name: The item name of parameter
    :param str description: Parameter description, or ``None``.
    :param dict raw: All defined data of the item
    :param str display_name: User-friendly name for display or documentation \
        purposes.  If ``displayName`` is not specified in RAML file, defaults \
        to ``name``.
    :param example: Example value for property, or ``None``.  Type of \
        ``example`` will match that of ``type``.
    :param default: Default value for property, or ``None``. Type of \
        ``default`` will match that of ``type``.
    :param int min_length: Parameter value's minimum number of characters, or \
        ``None``. Only applies when primative ``type`` is ``string``.
    :param int max_length: Parameter value's maximum number of characters, or \
        ``None``. Only applies when primative ``type`` is ``string``.
    :param int minimum: Parameter's minimum value, or ``None``.  Only applies \
        when primative ``type`` is ``integer`` or ``number``.
    :param int maximum: Parmeter's maximum value, or ``None``.  Only applies \
        when primative ``type`` is ``integer`` or ``number``.
    :param str type: Primative type of parameter. Defaults to ``string`` if \
        not set.
    :param list enum: Array of valid parameter values, or ``None``.  Only \
        applies when primative ``type`` is ``string``.
    :param bool repeat: If parameter can be repeated.  Defaults to ``False``.
    :param str pattern: A regular expression that parameter of type \
        ``string`` must match, or ``None`` if not set.
    :param str method: HTTP method for header, or ``None``
    :param bool required: If parameter is required. Defaults to ``False``.
    """
    name         = attr.ib(repr=False)
    display_name = attr.ib()
    raw          = attr.ib(repr=False,
                           validator=attr.validators.instance_of(dict))
    desc         = attr.ib(repr=False)
    example      = attr.ib(repr=False)
    default      = attr.ib(repr=False)
    min_length   = attr.ib(repr=False, validator=string_type_parameter)
    max_length   = attr.ib(repr=False, validator=string_type_parameter)
    minimum      = attr.ib(repr=False, validator=integer_number_type_parameter)
    maximum      = attr.ib(repr=False, validator=integer_number_type_parameter)
    config       = attr.ib(repr=False,
                           validator=attr.validators.instance_of(dict))
    errors       = attr.ib(repr=False)
    type         = attr.ib(repr=False, default="string", validator=header_type)
    enum         = attr.ib(repr=False, default=None,
                           validator=string_type_parameter)
    repeat       = attr.ib(repr=False, default=False)
    pattern      = attr.ib(repr=False, default=None,
                           validator=string_type_parameter)
    method       = attr.ib(repr=False, default=None)
    required     = attr.ib(repr=False, default=False)

    @property
    def description(self):
        if self.desc:
            return Content(self.desc)
        return None

    def _inherit_type_properties(self, inherited_param):
        params = NAMED_PARAMS + ["method"]
        for param in inherited_param:
            for n in params:
                attr = getattr(self, n, None)
                if attr is None:
                    attr = getattr(param, n, None)
                    setattr(self, n, attr)


@attr.s
class Body(object):
    """
    Body of the request/response.

    :param str mime_type: Accepted MIME media types for the body of the \
        request/response.
    :param dict raw: All defined data of the item
    :param dict schema: Body schema definition, or ``None`` if not set. \
        Can not be set if ``mime_type`` is ``multipart/form-data`` \
        or ``application/x-www-form-urlencoded``
    :param dict example: Example of schema, or ``None`` if not set. \
        Can not be set if ``mime_type`` is ``multipart/form-data`` \
        or ``application/x-www-form-urlencoded``
    :param dict form_params: Form parameters accepted in the body.
        Must be set if ``mime_type`` is ``multipart/form-data`` or \
        ``application/x-www-form-urlencoded``.  Can not be used when \
        schema and/or example is defined.
    """
    mime_type   = attr.ib(init=True, validator=body_mime_type)
    raw         = attr.ib(repr=False, init=True,
                          validator=attr.validators.instance_of(dict))
    schema      = attr.ib(repr=False, validator=body_schema)
    example     = attr.ib(repr=False, validator=body_example)
    form_params = attr.ib(repr=False, validator=body_form)
    config      = attr.ib(repr=False,
                          validator=attr.validators.instance_of(dict))
    errors      = attr.ib(repr=False)

    def _inherit_type_properties(self, inherited_param):
        body_params = ["schema", "example", "form_params"]
        for param in inherited_param:
            if param.mime_type != self.mime_type:
                continue
            for n in body_params:
                attr = getattr(self, n, None)
                if attr is None:
                    attr = getattr(param, n, None)
                    setattr(self, n, attr)


@attr.s
class Response(object):
    """
    Expected response parameters.

    :param int code: HTTP response code.
    :param dict raw: All defined data of item.
    :param str description: Response description, or ``None``.
    :param list headers: List of :py:class:`.Header` objects, or ``None``.
    :param list body: List of :py:class:`.Body` objects, or ``None``.
    :param str method: HTTP request method associated with response.
    """
    code     = attr.ib(validator=response_code)
    raw      = attr.ib(repr=False, init=True,
                       validator=attr.validators.instance_of(dict))
    desc     = attr.ib(repr=False)
    headers  = attr.ib(repr=False)
    body     = attr.ib(repr=False)
    config   = attr.ib(repr=False,
                       validator=attr.validators.instance_of(dict))
    errors   = attr.ib(repr=False)
    method   = attr.ib(default=None)

    @property
    def description(self):
        if self.desc:
            return Content(self.desc)
        return None

    def _inherit_type_properties(self, inherited_param):
        for param in inherited_param:
            for n in NAMED_PARAMS:
                attr = getattr(self, n, None)
                if attr is None:
                    attr = getattr(param, n, None)
                    setattr(self, n, attr)


@attr.s
class SecurityScheme(object):
    """
    Security scheme definition.

    :param str name: Name of security scheme
    :param dict raw: All defined data of item
    :param str type: Type of authentication
    :param dict described_by: :py:class:`.Header` s, :py:class:`.Response` s, \
        :py:class:`.QueryParameter` s, etc that is needed/can be expected \
        when using security scheme.
    :param str description: Description of security scheme
    :param dict settings: Security schema-specific information
    """
    name          = attr.ib()
    raw           = attr.ib(repr=False, init=True,
                            validator=attr.validators.instance_of(dict))
    type          = attr.ib(repr=False)
    described_by  = attr.ib(repr=False)
    desc          = attr.ib(repr=False)
    settings      = attr.ib(repr=False, validator=defined_sec_scheme_settings)
    config        = attr.ib(repr=False)
    errors        = attr.ib(repr=False)

    @property
    def description(self):
        if self.desc:
            return Content(self.desc)
        return None



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

from __future__ import absolute_import, division, print_function


from six import itervalues, iterkeys

from .parameters import SecurityScheme
from .utils import _get_scheme
# functions used to parse the same shit in parser.py


# resource, resource type
def security_schemes(secured, root):
    """Set resource's assigned security scheme objects."""
    if secured:
        secured_objs = []
        for item in secured:
            assigned_scheme = _get_scheme(item, root)
            if assigned_scheme:
                raw_data = list(itervalues(assigned_scheme))[0]
                scheme = SecurityScheme(
                    name=list(iterkeys(assigned_scheme))[0],
                    raw=raw_data,
                    type=raw_data.get("type"),
                    described_by=raw_data.get("describedBy"),
                    desc=raw_data.get("description"),
                    settings=raw_data.get("settings"),
                    config=root.config,
                    errors=root.errors
                )
                secured_objs.append(scheme)
        return secured_objs
    return None



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

from __future__ import absolute_import, division, print_function


import json
import logging
import os
import re
import sys

try:
    from collections import OrderedDict
except ImportError:  # NOCOV
    from ordereddict import OrderedDict

from six import iterkeys, iteritems
import xmltodict

from .parameters import (
    Body, URIParameter, Header, FormParameter, QueryParameter
)

PYVER = sys.version_info[:3]

if PYVER == (2, 7, 9) or PYVER == (3, 4, 3):  # NOCOV
    import six.moves.urllib.request as urllib
    import six.moves.urllib.error as urllib_error
    URLLIB = True
    SECURE_DOWNLOAD = True
else:
    try:  # NOCOV
        import requests
        URLLIB = False
        SECURE_DOWNLOAD = True
    except ImportError:
        import six.moves.urllib.request as urllib
        import six.moves.urllib.error as urllib_error
        URLLIB = True
        SECURE_DOWNLOAD = False

from .errors import MediaTypeError


IANA_URL = "https://www.iana.org/assignments/media-types/media-types.xml"


def load_schema(data):
    """
    Load Schema/Example data depending on its type (JSON, XML).

    If error in parsing as JSON and XML, just returns unloaded data.

    :param str data: schema/example data
    """
    try:
        return json.loads(data)
    except Exception:  # POKEMON!
        pass

    try:
        return xmltodict.parse(data)
    except Exception:  # GOTTA CATCH THEM ALL
        pass

    return data


def setup_logger(key):
    """General logger"""
    log = logging.getLogger(__name__)
    log.setLevel(logging.DEBUG)
    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG)
    msg = "{key} - %(levelname)s - %(message)s".format(key=key)
    formatter = logging.Formatter(msg)
    console.setFormatter(formatter)

    log.addHandler(console)
    return log


def _requests_download(url):
    """Download a URL using ``requests`` library"""
    try:
        response = requests.get(url)
        return response.text
    except requests.exceptions.RequestException as e:
        msg = "Error downloading from {0}: {1}".format(url, e)
        raise MediaTypeError(msg)


def _urllib_download(url):
    """Download a URL using ``urllib`` library"""
    try:
        response = urllib.urlopen(url)
    except urllib_error.URLError as e:
        msg = "Error downloading from {0}: {1}".format(url, e)
        raise MediaTypeError(msg)
    return response.read()


def download_url(url):
    """
    General download function, given a URL.

    If running 2.7.8 or earlier, or 3.4.2 or earlier, then use
    ``requests`` if it's installed.  Otherwise, use ``urllib``.
    """
    log = setup_logger("DOWNLOAD")
    if SECURE_DOWNLOAD and not URLLIB:
        return _requests_download(url)
    elif SECURE_DOWNLOAD and URLLIB:
        return _urllib_download(url)
    msg = ("Downloading over HTTPS but can not verify the host's "
           "certificate.  To avoid this in the future, `pip install"
           " \"requests[security]\"`.")
    log.warn(msg)
    return _urllib_download(url)


def _xml_to_dict(response_text):
    """Parse XML response from IANA into a Python ``dict``."""
    try:
        return xmltodict.parse(response_text)
    except xmltodict.expat.ExpatError as e:
        msg = "Error parsing XML: {0}".format(e)
        raise MediaTypeError(msg)


def _extract_mime_types(registry):
    """
    Parse out MIME types from a defined registry (e.g. "application",
    "audio", etc).
    """
    mime_types = []
    records = registry.get("record", {})
    reg_name = registry.get("@id")
    for rec in records:
        mime = rec.get("file", {}).get("#text")
        if mime:
            mime_types.append(mime)
        else:
            mime = rec.get("name")
            if mime:
                hacked_mime = reg_name + "/" + mime
                mime_types.append(hacked_mime)
    return mime_types


def _parse_xml_data(xml_data):
    """Parse the given XML data."""
    registries = xml_data.get("registry", {}).get("registry")
    if not registries:
        msg = "No registries found to parse."
        raise MediaTypeError(msg)
    if len(registries) is not 9:
        msg = ("Expected 9 registries but parsed "
               "{0}".format(len(registries)))
        raise MediaTypeError(msg)
    all_mime_types = []
    for registry in registries:
        mime_types = _extract_mime_types(registry)
        all_mime_types.extend(mime_types)

    return all_mime_types


def _save_updated_mime_types(output_file, mime_types):
    """Save the updated MIME Media types within the package."""
    with open(output_file, "w") as f:
        json.dump(mime_types, f)


def update_mime_types():
    """
    Update MIME Media Types from IANA.  Requires internet connection.
    """
    log = setup_logger("UPDATE")

    log.debug("Getting XML data from IANA")
    raw_data = download_url(IANA_URL)
    log.debug("Data received; parsing...")
    xml_data = _xml_to_dict(raw_data)
    mime_types = _parse_xml_data(xml_data)

    current_dir = os.path.dirname(os.path.realpath(__file__))
    data_dir = os.path.join(current_dir, "data")
    output_file = os.path.join(data_dir, "supported_mime_types.json")

    _save_updated_mime_types(output_file, mime_types)

    log.debug("Done! Supported IANA MIME media types have been updated.")


def _resource_type_lookup(assigned, root):
    """
    Returns ``ResourceType`` object

    :param str assigned: The string name of the assigned resource type
    :param root: RAML root object
    """
    res_types = root.resource_types
    if res_types:
        res_type_obj = [r for r in res_types if r.name == assigned]
        if res_type_obj:
            return res_type_obj[0]


#####
# Helper methods
#####

# general
def _get(data, item, default=None):
    """
    Helper function to catch empty mappings in RAML. If item is optional
    but not in the data, or data is ``None``, the default value is returned.

    :param data: RAML data
    :param str item: RAML key
    :param default: default value if item is not in dict
    :param bool optional: If RAML item is optional or needs to be defined
    :ret: value for RAML key
    """
    try:
        return data.get(item, default)
    except AttributeError:
        return default


def _create_base_param_obj(attribute_data, param_obj, config, errors, **kw):
    """Helper function to create a BaseParameter object"""
    objects = []

    for key, value in list(iteritems(attribute_data)):
        if param_obj is URIParameter:
            required = _get(value, "required", default=True)
        else:
            required = _get(value, "required", default=False)
        kwargs = dict(
            name=key,
            raw={key: value},
            desc=_get(value, "description"),
            display_name=_get(value, "displayName", key),
            min_length=_get(value, "minLength"),
            max_length=_get(value, "maxLength"),
            minimum=_get(value, "minimum"),
            maximum=_get(value, "maximum"),
            default=_get(value, "default"),
            enum=_get(value, "enum"),
            example=_get(value, "example"),
            required=required,
            repeat=_get(value, "repeat", False),
            pattern=_get(value, "pattern"),
            type=_get(value, "type", "string"),
            config=config,
            errors=errors
        )
        if param_obj is Header:
            kwargs["method"] = _get(kw, "method")

        item = param_obj(**kwargs)
        objects.append(item)

    return objects or None


# used for traits & resource nodes
def _map_param_unparsed_str_obj(param):
    return {
        "queryParameters": QueryParameter,
        "uriParameters": URIParameter,
        "formParameters": FormParameter,
        "baseUriParameters": URIParameter,
        "headers": Header
    }[param]


# create_resource_types
def _get_union(resource, method, inherited):
    union = {}
    for key, value in list(iteritems(inherited)):
        if resource.get(method) is not None:
            if key not in list(iterkeys(resource.get(method, {}))):
                union[key] = value
            else:
                resource_values = resource.get(method, {}).get(key)
                inherited_values = inherited.get(key, {})
                union[key] = dict(list(iteritems(resource_values)) +
                                  list(iteritems(inherited_values)))
    if resource.get(method) is not None:
        for key, value in list(iteritems(resource.get(method, {}))):
            if key not in list(iterkeys(inherited)):
                union[key] = value
    return union


def __is_scalar(item):
    scalar_props = [
        "type", "enum", "pattern", "minLength", "maxLength",
        "minimum", "maximum", "example", "repeat", "required",
        "default", "description", "usage", "schema", "example",
        "displayName"
    ]
    return item in scalar_props


def __get_sets(child, parent):
    child_keys = []
    parent_keys = []
    if child:
        child_keys = list(iterkeys(child))
    if parent:
        parent_keys = list(iterkeys(parent))
    child_diff = list(set(child_keys) - set(parent_keys))
    parent_diff = list(set(parent_keys) - set(child_keys))
    intersection = list(set(child_keys).intersection(parent_keys))
    opt_inters = [i for i in child_keys if str(i) + "?" in parent_keys]
    intersection = intersection + opt_inters

    return child, parent, child_diff, parent_diff, intersection


def _get_data_union(child, parent):
    # takes child data and parent data and merges them
    # with preference to child data overwriting parent data
    # confession: had to look up set theory!
    # FIXME: should bring this over from config, not hard code
    methods = [
        'get', 'post', 'put', 'delete', 'patch', 'head', 'options',
        'trace', 'connect', 'get?', 'post?', 'put?', 'delete?', 'patch?',
        'head?', 'options?', 'trace?', 'connect?'
    ]
    union = {}
    child, parent, c_diff, p_diff, inters = __get_sets(child, parent)

    for i in c_diff:
        union[i] = child.get(i)
    for i in p_diff:
        if i in methods and not i.endswith("?"):
                union[i] = parent.get(i)
        if i not in methods:
            union[i] = parent.get(i)
    for i in inters:
        if __is_scalar(i):
            union[i] = child.get(i)
        else:
            _child = child.get(i, {})
            _parent = parent.get(i, {})
            union[i] = _get_data_union(_child, _parent)
    return union


def _get_inherited_resource(res_name, resource_types):
    for resource in resource_types:
        if res_name == list(iterkeys(resource))[0]:
            return resource


def _get_res_type_attribute(res_data, method_data, item, default={}):
    method_level = _get(method_data, item, default)
    resource_level = _get(res_data, item, default)
    return method_level, resource_level


def _get_inherited_type_params(data, attribute, params, resource_types):
    inherited = _get_inherited_resource(data.get("type"), resource_types)
    inherited = inherited.get(data.get("type"))
    inherited_params = inherited.get(attribute, {})

    return dict(list(iteritems(params)) +
                list(iteritems(inherited_params)))


def _get_inherited_item(items, item_name, res_types, meth_, _data):
    inherited = _get_inherited_resource(_data.get("type"), res_types)
    resource = inherited.get(_data.get("type"))
    res_level = resource.get(meth_, {}).get(item_name, {})

    method_ = resource.get(meth_, {})
    method_level = method_.get(item_name, {})
    items = dict(
        list(iteritems(items)) +
        list(iteritems(res_level)) +
        list(iteritems(method_level))
    )
    return items


def _get_attribute_dict(data, item, v):
    resource_level = _get(v, item, {})
    method_level = _get(data, item, {})
    return dict(list(iteritems(resource_level)) +
                list(iteritems(method_level)))


def _map_attr(attribute):
    return {
        "mediaType": "media_type",
        "protocols": "protocols",
        "headers": "headers",
        "body": "body",
        "responses": "responses",
        "uriParameters": "uri_params",
        "baseUriParameters": "base_uri_params",
        "queryParameters": "query_params",
        "formParameters": "form_params",
        "description": "description"
    }[attribute]


# create_node
def _get_method(attribute, method, raw_data):
    """Returns ``attribute`` defined at the method level, or ``None``."""
    # if method is not None:  # must explicitly say `not None`
    #     get_attribute = raw_data.get(method, {})
    #     if get_attribute is not None:
    #         return get_attribute.get(attribute, {})
    # return {}
    # if method is not None:
    ret = _get(raw_data, method, {})
    ret = _get(ret, attribute, {})
    return ret


def _get_resource(attribute, raw_data):
    """Returns ``attribute`` defined at the resource level, or ``None``."""
    return raw_data.get(attribute, {})


def _get_parent(attribute, parent):
    if parent:
        return getattr(parent, attribute, {})
    return {}


# needs/uses parsed raml data
def _get_resource_type(attribute, root, type_, method):
    """Returns ``attribute`` defined in the resource type, or ``None``."""
    if type_ and root.resource_types:
        types = root.resource_types
        r_type = [r for r in types if r.name == type_]
        r_type = [r for r in r_type if r.method == method]
        if r_type:
            if hasattr(r_type[0], attribute):
                if getattr(r_type[0], attribute) is not None:
                    return getattr(r_type[0], attribute)
    return []


def _get_trait(attribute, root, is_):
    """Returns ``attribute`` defined in a trait, or ``None``."""

    if is_:
        traits = root.traits
        if traits:
            trait_objs = []
            for i in is_:
                trait = [t for t in traits if t.name == i]
                if trait:
                    if hasattr(trait[0], attribute):
                        if getattr(trait[0], attribute) is not None:
                            trait_objs.extend(getattr(trait[0], attribute))
            return trait_objs
    return []


def _get_scheme(item, root):
    schemes = root.raw.get("securitySchemes", [])
    for s in schemes:
        if isinstance(item, str):
            if item == list(iterkeys(s))[0]:
                return s
        elif isinstance(item, dict):
            if list(iterkeys(item))[0] == list(iterkeys(s))[0]:
                return s


def _get_attribute(attribute, method, raw_data):
    method_level = _get_method(attribute, method, raw_data)
    resource_level = _get_resource(attribute, raw_data)
    return OrderedDict(list(iteritems(method_level)) +
                       list(iteritems(resource_level)))


def _get_inherited_attribute(attribute, root, type_, method, is_):
    type_objects = _get_resource_type(attribute, root, type_, method)
    trait_objects = _get_trait(attribute, root, is_)
    return type_objects + trait_objects


# TODO: refactor - this ain't pretty
def _remove_duplicates(inherit_params, resource_params):
    ret = []
    if isinstance(resource_params[0], Body):
        _params = [p.mime_type for p in resource_params]
    else:
        _params = [p.name for p in resource_params]

    for p in inherit_params:
        if isinstance(p, Body):
            if p.mime_type not in _params:
                ret.append(p)
        else:
            if p.name not in _params:
                ret.append(p)
    ret.extend(resource_params)
    return ret or None


def _map_inheritance(nodetype):
    return {
        "traits": __trait,
        "types": __resource_type,
        "method": __method,
        "resource": __resource,
        "parent": __parent,
        "root": __root
    }[nodetype]


def __trait(item, **kwargs):
    root = kwargs.get("root")
    is_ = kwargs.get("is_")
    return _get_trait(item, root, is_)


def __resource_type(item, **kwargs):
    root = kwargs.get("root")
    type_ = kwargs.get("type_")
    method = kwargs.get("method")
    item = _map_attr(item)
    return _get_resource_type(item, root, type_, method)


def __method(item, **kwargs):
    method = kwargs.get("method")
    data = kwargs.get("data")
    return _get_method(item, method, data)


def __resource(item, **kwargs):
    data = kwargs.get("data")
    return _get_resource(item, data)


def __parent(item, **kwargs):
    parent = kwargs.get("parent")
    return _get_parent(item, parent)


def __root(item, **kwargs):
    root = kwargs.get("root")
    item = _map_attr(item)
    return getattr(root, item, None)


def get_inherited(item, inherit_from=[], **kwargs):
    ret = {}
    for nodetype in inherit_from:
        inherit_func = _map_inheritance(nodetype)
        inherited = inherit_func(item, **kwargs)
        ret[nodetype] = inherited
    return ret


#####
# set uri, form, query, header objects for traits
#####

def set_param_object(param_data, param_str, root):
    params = _get(param_data, param_str, {})
    param_obj = _map_param_unparsed_str_obj(param_str)
    return _create_base_param_obj(params,
                                  param_obj,
                                  root.config,
                                  root.errors)


#####
# set query, form, uri params for resource nodes
#####

# <--[uri]-->
def __create_params(unparsed, parsed, method, raw_data, root, cls, type_, is_):
    _params = _get_attribute(unparsed, method, raw_data)
    param_objs = _get_inherited_attribute(parsed, root, type_,
                                          method, is_)
    params = _create_base_param_obj(_params, cls, root.config, root.errors)
    return params, param_objs


def _create_uri_params(unparsed, parsed, root_params, root, type_, is_,
                       method, raw_data, parent):
    params, param_objs = __create_params(unparsed, parsed, method, raw_data,
                                         root, URIParameter, type_, is_)

    if params:
        param_objs.extend(params)
    if parent and parent.uri_params:
        param_objs.extend(parent.uri_params)
    if root_params:
        param_names = [p.name for p in param_objs]
        _params = [p for p in root_params if p.name not in param_names]
        param_objs.extend(_params)
    return param_objs or None
# <--[uri]-->


# <--[query, base uri, form]-->
def _check_already_exists(param, ret_list):
    if isinstance(param, Body):
        param_name_list = [p.mime_type for p in ret_list]
        if param.mime_type not in param_name_list:
            ret_list.append(param)
            param_name_list.append(param.mime_type)

    else:
        param_name_list = [p.name for p in ret_list]
        if param.name not in param_name_list:
            ret_list.append(param)
            param_name_list.append(param.name)
    return ret_list


# TODO: refactor - this ain't pretty
def __remove_duplicates(to_clean):
    # order: resource, inherited, parent, root
    ret = []

    for param_set in to_clean:
        if param_set:
            for p in param_set:
                ret = _check_already_exists(p, ret)
    return ret or None


def _map_parsed_str(parsed):
    name = parsed.split("_")[:-1]
    name.append("parameters")
    name = [n.capitalize() for n in name]
    name = "".join(name)
    return name[0].lower() + name[1:]


def set_params(data, param_str, root, method, inherit=False, **kw):
    params, param_objs, parent_params, root_params = [], [], [], []

    unparsed = _map_parsed_str(param_str)
    param_obj = _map_param_unparsed_str_obj(unparsed)
    _params = _get_attribute(unparsed, method, data)

    params = _create_base_param_obj(_params, param_obj,
                                    root.config, root.errors)
    if params is None:
        params = []

    # inherited objects
    if inherit:
        type_ = kw.get("type_")
        is_ = kw.get("is_")
        param_objs = _get_inherited_attribute(param_str, root, type_,
                                              method, is_)

    # parent objects
    parent = kw.get("parent")
    if parent:
        parent_params = getattr(parent, param_str, [])

    # root objects
    root = kw.get("root_params")
    if root:
        param_names = [p.name for p in param_objs]
        root_params = [p for p in root if p.name not in param_names]

    to_clean = (params, param_objs, parent_params, root_params)

    return __remove_duplicates(to_clean)
# <--[query, base uri, form]-->


# preserve order of URI and Base URI parameters
# used for RootNode, ResourceNode
def _preserve_uri_order(path, param_objs, config, errors, declared=[]):
    # if this is hit, RAML shouldn't be valid anyways.
    if isinstance(path, list):
        path = path[0]

    sorted_params = []
    pattern = "\{(.*?)\}"
    params = re.findall(pattern, path)
    if not param_objs:
        param_objs = []
    # if there are URI parameters in the path but were not declared
    # inline, we should create them.
    # TODO: Probably shouldn't do it in this function, though...
    if len(params) > len(param_objs):
        if len(param_objs) > 0:
            param_names = [p.name for p in param_objs]
            missing = [p for p in params if p not in param_names]
        else:
            missing = params[::]
        # exclude any (base)uri params if already declared
        missing = [p for p in missing if p not in declared]
        for m in missing:
            # no need to create a URI param for version
            if m == "version":
                continue
            data = {"type": "string"}
            _param = URIParameter(name=m,
                                  raw={m: data},
                                  required=True,
                                  display_name=m,
                                  desc=_get(data, "description"),
                                  min_length=_get(data, "minLength"),
                                  max_length=_get(data, "maxLength"),
                                  minimum=_get(data, "minimum"),
                                  maximum=_get(data, "maximum"),
                                  default=_get(data, "default"),
                                  enum=_get(data, "enum"),
                                  example=_get(data, "example"),
                                  repeat=_get(data, "repeat", False),
                                  pattern=_get(data, "pattern"),
                                  type=_get(data, "type", "string"),
                                  config=config,
                                  errors=errors)
            param_objs.append(_param)
    for p in params:
        _param = [i for i in param_objs if i.name == p]
        if _param:
            sorted_params.append(_param[0])
    return sorted_params or None



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

import os

import pytest

from ramlfications import parse, load, loads, validate
from ramlfications.raml import RootNode
from ramlfications.errors import LoadRAMLError

from .base import EXAMPLES


@pytest.fixture(scope="session")
def raml():
    return os.path.join(EXAMPLES + "complete-valid-example.raml")


@pytest.fixture(scope="session")
def raml_string():
    return """#%RAML 0.8
---
title: GitHub API
version: v3
baseUri: https://api.github.com/
"""


def test_parse(raml):
    config = os.path.join(EXAMPLES + "test-config.ini")
    result = parse(raml, config)
    assert result
    assert isinstance(result, RootNode)


def test_parse_nonexistant_file():
    raml_file = "/tmp/non-existant-raml-file.raml"
    with pytest.raises(LoadRAMLError):
        parse(raml_file)


def test_load(raml):
    result = load(raml)
    assert result
    assert isinstance(result, dict)


def test_load_nonexistant_file():
    raml_file = "/tmp/non-existant-raml-file.raml"
    with pytest.raises(LoadRAMLError):
        parse(raml_file)


def test_loads(raml_string):
    result = loads(raml_string)
    assert result
    assert isinstance(result, dict)


def test_validate(raml):
    result = validate(raml)
    assert result is None


def test_validate_nonexistant_file():
    raml_file = "/tmp/non-existant-raml-file.raml"
    with pytest.raises(LoadRAMLError):
        validate(raml_file)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB
from __future__ import absolute_import, division, print_function


import os

from tests.base import JSONREF


# For test_loader.py::test_load_file
load_file_expected_data = {
    'external': {
        'propertyA': 'valueA',
        'propertyB': 'valueB'
    },
    'foo': {
        'foo': 'FooBar',
        'bar': 'BarBaz'
    },
    'title': 'GitHub API Demo - Includes',
    'version': 'v3'
}

# For test_loader.py::test_load_file_with_nested_includes
load_file_with_nested_includes_expected = {
    'include_one': {
        'external': {
            'propertyA': 'valueA',
            'propertyB': 'valueB'
        },
        'foo': {
            'foo': 'FooBar',
            'bar': 'BarBaz'
        },
        'not_yaml': "This is just a string.\n",
    },
    'title': 'GitHub API Demo - Includes',
    'version': 'v3'
}

# For test_loader.py::test_include_json
include_json_expected = {
    "title": "Sample API Demo - JSON Includes",
    "version": "v1",
    "baseUri": "http://json.example.com",
    "schemas": [{
        "json": {
            "name": "foo",
            "false": True
        },
    }],
    "/foo": {
        "displayName": "foo resource"
    }
}

# For test_loader.py::test_include_xsd
include_xsd_raw = """<?xml version="1.0" encoding="UTF-8"?>
<root>
   <false>true</false>
   <name>foo</name>
</root>
"""

include_xsd_expected = {
    "title": "Sample API Demo - XSD Includes",
    "version": "v1",
    "baseUri": "http://xml.example.com",
    "schemas": [{
        "xml": include_xsd_raw,
    }],
    "/foo": {
        "displayName": "foo resource",
    },
}

# For test_loader.py::test_include_markdown
include_markdown_raw = """## Foo

*Bacon ipsum dolor* amet pork belly _doner_ rump brisket. [Cupim jerky \
shoulder][0] ball tip, jowl bacon meatloaf shank kielbasa turducken corned \
beef beef turkey porchetta.

### Doner meatball pork belly andouille drumstick sirloin

Porchetta picanha tail sirloin kielbasa, pig meatball short ribs drumstick \
jowl. Brisket swine spare ribs picanha t-bone. Ball tip beef tenderloin jowl \
doner andouille cupim meatball. Porchetta hamburger filet mignon jerky flank, \
meatball salami beef cow venison tail ball tip pork belly.

[0]: https://baconipsum.com/?paras=1&type=all-meat&start-with-lorem=1
"""

include_markdown_expected = {
    "title": "Sample API Demo - Markdown Includes",
    "version": "v1",
    "baseUri": "http://markdown.example.com",
    "/foo": {
        "displayName": "foo resource"
    },
    "documentation": [{
        "title": "example",
        "content": include_markdown_raw,
    }],
}

# For test_loader.py::test_jsonref_relative_empty_fragment
jsonref_relative_empty_fragment_expected = {
    "title": "Sample API Demo - JSON Includes",
    "version": "v1",
    "baseUri": "http://json.example.com",
    "schemas": [{
        "json": {
            "name": "foo",
            "false": True
        },
    }],
    "/foo": {
        "displayName": "foo resource"
    }
}

# For test_loader.py::test_json_ref_in_schema_relative_nonempty_fragment
jsonref_relative_nonempty_fragment_expected = {
    "title": "Sample API Demo - JSON Includes",
    "version": "v1",
    "baseUri": "http://json.example.com",
    "schemas": [{
        "json": {
            "first_name": "foo",
            "second_name": "bar"
        },
    }],
    "/foo": {
        "displayName": "foo resource"
    }
}

# For test_loader.py::test_jsonref_internal_fragment_reference
jsonref_internal_fragment_reference_expected = {
    "title": "Sample API Demo - JSON Includes",
    "version": "v1",
    "baseUri": "http://json.example.com",
    "schemas": [{
        "json": {
            "references":
            {
                "internal": "yes",
            },
            "name": "foo",
            "is_this_internal?": ["yes"],
        },
    }],
    "/foo": {
        "displayName": "foo resource"
    }
}

# For test_loader.py::test_jsonref_multiref_internal_fragments
jsonref_multiref_internal_fragments_expected = {
    "title": "Sample API Demo - JSON Includes",
    "version": "v1",
    "baseUri": "http://json.example.com",
    "schemas": [{
        "json": {
            "references":
            {
                "internal": "yes",
                "two": True,
            },
            "name": "foo",
            "is_this_internal?": "yes",
            "two_references": True,
        },
    }],
    "/foo": {
        "displayName": "foo resource",
    }
}


# For test_loader.py::test_jsonref_absolute_local_uri
json_ref_absolute_jsonfile = os.path.join(JSONREF, "jsonref_example.json")

json_ref_absolute_jsondump = {
    "second_name": "bar",
    "$ref": json_ref_absolute_jsonfile
}

json_ref_absolute_ramlfile = ("""#%RAML 0.8
title: Sample API Demo - JSON Includes
version: v1
schemas:
    - json: !include {json_file}
baseUri: http://json.example.com
/foo:
  displayName: foo resource
""")


json_ref_absolute_expected = {
    "title": "Sample API Demo - JSON Includes",
    "version": "v1",
    "baseUri": "http://json.example.com",
    "schemas": [{
        "json": {
            "name": "foo",
            "false": True
        },
    }],
    "/foo": {
        "displayName": "foo resource"
    }
}


# for test_loader.py::test_jsonref_relative_local_uri
jsonref_relative_local_expected = {
    '$schema': 'http://json-schema.org/draft-03/schema',
    'properties': {
        'album_type': {
            'description': ("The type of the album: one of 'album', "
                            "'single', or 'compilation'."),
            'type': 'string'
        },
        'artists': {
            'description': ("The artists of the album. Each artist "
                            "object includes a link in href to more "
                            "detailed information about the artist."),
            'items': [{
                'popularity': {
                    'type': 'integer',
                    'description': ("The popularity of the artist. The "
                                    "value will be between 0 and 100, "
                                    "with 100 being the most popular. "
                                    "The artist's popularity is "
                                    "calculated from the popularity of "
                                    "all the artist's tracks.")
                },
                'type': {
                    'type': 'string',
                    'description': "The object type: 'artist'"
                },
                'name': {
                    'type': 'string',
                    'description': 'The name of the artist.'
                },
                'uri': {
                    'type': 'string',
                    'description': 'The Spotify URI for the artist.'
                }
            }],
            'type': 'array'
        }
    },
    'type': 'object'
}

# for test_loader.py::test_jsonref_relative_local_uri_includes
jsonref_relative_local_includes_expected = {
    '$schema': 'http://json-schema.org/draft-03/schema',
    'properties': {
        'album_type': {
            'description': ("The type of the album: one of 'album', "
                            "'single', or 'compilation'."),
            'type': u'string'
        },
        'artists': {
            'description': ("The artists of the album. Each artist "
                            "object includes a link in href to more "
                            "detailed information about the artist."),
            'items': [{
                'genres': {
                    'items': {
                        'type': 'string'
                    },
                    'type': 'array',
                    'description': ("A list of the genres the artist is "
                                    "associated with. For example: 'Prog "
                                    "Rock', 'Post-Grunge'. (If not yet "
                                    "classified, the array is empty.)")
                },
                'name': {
                    'type': 'string',
                    'description': 'The name of the artist.'
                },
                'external_urls': {
                    '': [{
                        'type': 'string'
                    }],
                    'type': 'object',
                    'description': 'Known external URLs for this artist.'
                },
                'popularity': {
                    'type': 'integer',
                    'description': ("The popularity of the artist. The "
                                    "value will be between 0 and 100, "
                                    "with 100 being the most popular. "
                                    "The artist's popularity is "
                                    "calculated from the popularity of "
                                    "all the artist's tracks.")
                },
                'uri': {
                    'type': 'string',
                    'description': 'The Spotify URI for the artist.'
                },
                'href': {
                    'type': 'string',
                    'description': ("A link to the Web API endpoint "
                                    "providing full details of the "
                                    "artist.")
                },
                'images': {
                    'items': {
                        'url': {
                            'type': 'string',
                            'description': ("The source URL of the "
                                            "image.")
                        },
                        'width': {
                            'type': 'integer',
                            'description': ("The image width in pixels. "
                                            "If unknown: null or not "
                                            "returned.")
                        },
                        'height': {
                            'type': 'integer',
                            'description': ("The image height in "
                                            "pixels. If unknown: "
                                            "null or not returned.")
                        }
                    },
                    'type': 'array',
                    'description': ("Images of the artist in "
                                    "various sizes, widest first.")
                },
                'type': {
                    'type': 'string',
                    'description': ("The object type: 'artist'")
                },
                'id': {
                    'type': 'string',
                    'description': ("The Spotify ID for the artist.")
                }
            }],
            'type': 'array'
        }
    },
    'type': 'object'
}


# for test_loader.py::test_jsonref_remote_uri
jsonref_remote_uri_jsondump = {
    "$schema": "http://json-schema.org/draft-03/schema",
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "description": "The name of the artist."
        },
        "images": {
            "type": "array",
            "description": "Images associated with artist",
            "items": [{"$ref": "{url}"}]
        }
    }
}

jsonref_remote_uri_raml = ("""#%RAML 0.8
title: Sample API Demo - JSON Includes
version: v1
schemas:
    - json: !include {json_file}
baseUri: http://json.example.com
/foo:
  displayName: foo resource
""")


json_remote_uri_expected = {
    "title": "Sample API Demo - JSON Includes",
    "version": "v1",
    "baseUri": "http://json.example.com",
    "schemas": [{
        "json": {
            "$schema": "http://json-schema.org/draft-03/schema",
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                    "description": "The name of the artist."
                },
                "images": {
                    "type": "array",
                    "description": "Images associated with artist",
                    "items": [{
                        "height": {
                            "type": "integer",
                            "description": ("The image height in pixels. "
                                            "If unknown: null or not "
                                            "returned.")
                        },
                        "url": {
                            "type": "string",
                            "description": "The source URL of the image."
                        },
                        "width": {
                            "type": "integer",
                            "description": ("The image width in pixels. "
                                            "If unknown: null or not "
                                            "returned.")
                        }
                    }]
                }
            },
        }
    }],
    "/foo": {
        "displayName": "foo resource"
    }
}

jsonref_relative_local_file_expected = {
    "$schema": "http://json-schema.org/draft-03/schema",
    "type": "object",
    "properties": {
        "album_type": {
            "type": "string",
            "description": ("The type of the album: one of 'album', 'single', "
                            "or 'compilation'.")
        },
        "artists": {
            "type": "array",
            "description": ("The artists of the album. Each artist object "
                            "includes a link in href to more detailed "
                            "information about the artist."),
            "items": [{
                'popularity': {
                    'type': 'integer',
                    'description': ("The popularity of the artist. The "
                                    "value will be between 0 and 100, "
                                    "with 100 being the most popular. "
                                    "The artist's popularity is "
                                    "calculated from the popularity of "
                                    "all the artist's tracks.")
                },
                'type': {
                    'type': 'string',
                    'description': "The object type: 'artist'"
                },
                'name': {
                    'type': 'string',
                    'description': 'The name of the artist.'
                },
                'uri': {
                    'type': 'string',
                    'description': 'The Spotify URI for the artist.'
                }
            }]
        }
    }
}


# For test_loader.py::test_jsonref_absolute_local_uri_file
json_ref_absolute_jsonfile = os.path.join(JSONREF, "jsonref_example.json")
json_ref_absolute_jsonfile_file = "file://" + json_ref_absolute_jsonfile

json_ref_absolute_jsondump_file = {
    "second_name": "bar",
    "$ref": json_ref_absolute_jsonfile_file
}



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB
from __future__ import absolute_import, division, print_function

import os

import pytest

from six import iteritems

from ramlfications.config import setup_config
from ramlfications.config import (
    AUTH_SCHEMES, HTTP_RESP_CODES, MEDIA_TYPES, PROTOCOLS, HTTP_METHODS,
    RAML_VERSIONS, PRIM_TYPES
)

from .base import EXAMPLES


@pytest.fixture
def config():
    return os.path.join(EXAMPLES + "test-config.ini")


@pytest.fixture
def basic_config():
    basic_conf = {
        "auth_schemes": AUTH_SCHEMES,
        "resp_codes": HTTP_RESP_CODES,
        "media_types": MEDIA_TYPES,
        "protocols": PROTOCOLS,
        "http_methods": HTTP_METHODS,
        "raml_versions": RAML_VERSIONS,
        "prim_types": PRIM_TYPES,
        "validate": True,
    }
    optional = [m + "?" for m in basic_conf["http_methods"]]
    basic_conf["http_optional"] = optional + basic_conf["http_methods"]
    return basic_conf


def _clean(a_list):
    return sorted(list(set(a_list)))


def _dict_equal(dict1, dict2):
    for k, v1 in list(iteritems(dict1)):
        assert k in dict2
        v2 = dict2[k]
        assert v1, v2
    return True


def test_no_config_file_given(basic_config):
    parsed_config = setup_config(config_file=None)
    assert parsed_config == basic_config


def test_config_file(config, basic_config):
    bc = basic_config  # lazy
    bc["resp_codes"] = _clean(bc["resp_codes"] + [420, 421, 422])
    bc["auth_schemes"] = _clean(
        bc["auth_schemes"] + ["oauth_3_0", "oauth_4_0"]
    )
    bc["media_types"] = _clean(
        bc["media_types"] + ["application/vnd.github.v3", "foo/bar"]
    )
    bc["protocols"] = _clean(bc["protocols"] + ["FTP"])
    bc["raml_versions"] = _clean(bc["raml_versions"] + ["0.8"])
    bc["production"] = 'True'
    bc["validate"] = 'True'

    parsed_config = setup_config(config)

    assert _dict_equal(bc, parsed_config)


def test_no_config_file_found():
    config_file = "/tmp/a-nonexistent-RAML-config-file.ini"
    with pytest.raises(IOError) as e:
        setup_config(config_file)

    msg = ("No such file or directory: '{0}'".format(config_file),)
    assert e.value.args == msg



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB
import sys

if sys.version_info[0] == 2:
    from io import open

import json
import os
import tempfile

from mock import Mock, patch
import pytest
import xmltodict

from ramlfications import utils

from .base import UPDATE


@pytest.fixture(scope="session")
def downloaded_xml():
    return os.path.join(UPDATE, "iana_mime_media_types.xml")


@pytest.fixture(scope="session")
def invalid_xml():
    return os.path.join(UPDATE, "invalid_iana_download.xml")


@pytest.fixture(scope="session")
def no_data_xml():
    return os.path.join(UPDATE, "no_data.xml")


@pytest.fixture(scope="session")
def expected_data():
    expected_json = os.path.join(UPDATE, "expected_mime_types.json")
    with open(expected_json, "r", encoding="UTF-8") as f:
        return json.load(f)


@pytest.fixture(scope="session")
def parsed_xml(downloaded_xml):
    with open(downloaded_xml, "r", encoding="UTF-8") as f:
        data = f.read()
        return xmltodict.parse(data)


def test_xml_to_dict(downloaded_xml):
    with open(downloaded_xml, "r", encoding="UTF-8") as f:
        data = f.read()
        xml_data = utils._xml_to_dict(data)
        assert xml_data is not None
        assert isinstance(xml_data, dict)


def test_xml_to_dict_no_data(no_data_xml):
    with pytest.raises(utils.MediaTypeError) as e:
        with open(no_data_xml, "r", encoding="UTF-8") as f:
            data = f.read()
            utils._xml_to_dict(data)

    msg = "Error parsing XML: "
    assert msg in e.value.args[0]


def test_xml_to_dict_invalid(invalid_xml):
    with pytest.raises(utils.MediaTypeError) as e:
        with open(invalid_xml, "r", encoding="UTF-8") as f:
            data = f.read()
            utils._xml_to_dict(data)

    msg = "Error parsing XML: "
    assert msg in e.value.args[0]


def test_parse_xml_data(parsed_xml, expected_data):
    result = utils._parse_xml_data(parsed_xml)

    assert result == expected_data
    assert len(result) == len(expected_data)


@pytest.fixture(scope="session")
def incorrect_registry_count():
    xml_file = os.path.join(UPDATE, "unexpected_registry_count.xml")
    with open(xml_file, "r", encoding="UTF-8") as f:
        data = f.read()
        return xmltodict.parse(data)


def test_parse_xml_data_incorrect_reg(incorrect_registry_count):
    with pytest.raises(utils.MediaTypeError) as e:
        utils._parse_xml_data(incorrect_registry_count)

    msg = ("Expected 9 registries but parsed 2",)
    assert e.value.args == msg


@pytest.fixture(scope="session")
def no_registries():
    xml_file = os.path.join(UPDATE, "no_registries.xml")
    with open(xml_file, "r", encoding="UTF-8") as f:
        data = f.read()
        return xmltodict.parse(data)


def test_parse_xml_data_no_reg(no_registries):
    with pytest.raises(utils.MediaTypeError) as e:
        utils._parse_xml_data(no_registries)

    msg = ("No registries found to parse.",)
    assert e.value.args == msg


def test_requests_download_xml(downloaded_xml):
    utils.requests = Mock()
    with open(downloaded_xml, "r", encoding="UTF-8") as xml:
        expected = xml.read()
        utils.requests.get.return_value.text = expected
        results = utils._requests_download(utils.IANA_URL)

        assert results == expected


def test_urllib_download(downloaded_xml):
    utils.urllib = Mock()
    with open(downloaded_xml, "r", encoding="UTF-8") as xml:
        utils.urllib.urlopen.return_value = xml
        results = utils._urllib_download(utils.IANA_URL)

    with open(downloaded_xml, "r", encoding="UTF-8") as xml:
        assert results == xml.read()


@patch("ramlfications.utils._parse_xml_data")
@patch("ramlfications.utils._xml_to_dict")
@patch("ramlfications.utils._save_updated_mime_types")
def test_insecure_download_urllib_flag(_a, _b, _c, mocker, monkeypatch):
    monkeypatch.setattr(utils, "SECURE_DOWNLOAD", False)
    monkeypatch.setattr(utils, "URLLIB", True)
    utils.requests = Mock()

    mocker.patch("ramlfications.utils._urllib_download")

    utils.update_mime_types()
    utils._urllib_download.assert_called_once()

    mocker.stopall()


@patch("ramlfications.utils._xml_to_dict")
@patch("ramlfications.utils._parse_xml_data")
@patch("ramlfications.utils._save_updated_mime_types")
def test_secure_download_requests_flag(_a, _b_, _c, mocker, monkeypatch):
    monkeypatch.setattr(utils, "SECURE_DOWNLOAD", True)
    monkeypatch.setattr(utils, "URLLIB", False)
    utils.urllib = Mock()

    mocker.patch("ramlfications.utils._requests_download")

    utils.update_mime_types()
    utils._requests_download.assert_called_once()

    mocker.stopall()


@patch("ramlfications.utils._xml_to_dict")
@patch("ramlfications.utils._parse_xml_data")
@patch("ramlfications.utils._requests_download")
@patch("ramlfications.utils._urllib_download")
@patch("ramlfications.utils._save_updated_mime_types")
def test_update_mime_types(_a, _b, _c, _d, _e, downloaded_xml):
    utils.requests = Mock()

    with open(downloaded_xml, "r", encoding="UTF-8") as raw_data:
        utils.update_mime_types()
        utils._requests_download.assert_called_once()
        utils._requests_download.return_value = raw_data.read()
        utils._xml_to_dict.assert_called_once()
        utils._parse_xml_data.assert_called_once()
        utils._save_updated_mime_types.assert_called_once()


def test_save_updated_mime_types():
    content = ["foo/bar", "bar/baz"]
    temp_output = tempfile.mkstemp()[1]
    utils._save_updated_mime_types(temp_output, content)

    result = open(temp_output, "r", encoding="UTF-8").read()
    result = json.loads(result)
    assert result == content

    os.remove(temp_output)



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB
import sys

if sys.version_info[0] == 2:
    from io import open

import os

import six

from .errors import LoadRAMLError
from .loader import RAMLLoader


def load_file(raml_file):
    try:
        with _get_raml_object(raml_file) as raml:
            return RAMLLoader().load(raml)
    except IOError as e:
        raise LoadRAMLError(e)


def load_string(raml_str):
    return RAMLLoader().load(raml_str)


def _get_raml_object(raml_file):
    """
    Returns a file object.
    """
    if raml_file is None:
        msg = "RAML file can not be 'None'."
        raise LoadRAMLError(msg)

    if isinstance(raml_file, six.text_type) or isinstance(
            raml_file, bytes):
        return open(os.path.abspath(raml_file), 'r', encoding="UTF-8")
    elif hasattr(raml_file, 'read'):
        return raml_file
    else:
        msg = ("Can not load object '{0}': Not a basestring type or "
               "file object".format(raml_file))
        raise LoadRAMLError(msg)



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB
from __future__ import absolute_import, division, print_function

import os

import pytest

from ramlfications import errors
from ramlfications import validate

from .base import VALIDATE


raises = pytest.raises(errors.InvalidRAMLError)


# Search a list of errors for a specific error
def _error_exists(error_list, error_type, error_msg):
    for e in error_list:
        if isinstance(e, error_type) and e.args == error_msg:
            return True

    return False


def load_raml(filename):
    return os.path.join(VALIDATE + filename)


def load_config(filename):
    return os.path.join(VALIDATE + filename)


def test_default_to_validate():
    # Should validate even though it's set to "false" in config
    raml = "tests/data/validate/invalid-protocols.raml"
    config = "tests/data/validate/validation-off.ini"
    with raises as e:
        validate(raml, config)
    msg = ("'FTP' not a valid protocol for a RAML-defined API.",)
    assert _error_exists(e.value.errors, errors.InvalidRootNodeError, msg)


def test_default_to_validate_no_config():
    # Should validate even though it's set to "false" in config
    raml = "tests/data/validate/invalid-protocols.raml"
    with raises as e:
        validate(raml)
    msg = ("'FTP' not a valid protocol for a RAML-defined API.",)
    assert _error_exists(e.value.errors, errors.InvalidRootNodeError, msg)


def test_invalid_root_protocols():
    raml = "tests/data/validate/invalid-protocols.raml"
    config = "tests/data/validate/valid-config.ini"
    with raises as e:
        validate(raml, config)
    msg = ("'FTP' not a valid protocol for a RAML-defined API.",)
    assert _error_exists(e.value.errors, errors.InvalidRootNodeError, msg)


def test_undefined_version():
    raml = load_raml("no-version.raml")
    config = load_config("valid-config.ini")
    parsed_raml = validate(raml, config)
    assert not hasattr(parsed_raml, "errors")


def test_invalid_version_base_uri():
    raml = load_raml("no-version-base-uri.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("RAML File's baseUri includes {version} parameter but no "
           "version is defined.",)
    assert _error_exists(e.value.errors, errors.InvalidRootNodeError, msg)


def test_undefined_base_uri_and_title():
    raml = load_raml("no-base-uri-no-title.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    assert len(e.value.errors) == 2
    assert isinstance(e.value.errors[0], errors.InvalidRootNodeError)
    assert isinstance(e.value.errors[1], errors.InvalidRootNodeError)


def test_invalid_base_uri_not_defined():
    raml = load_raml("no-base-uri.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("RAML File does not define the baseUri.",)
    assert _error_exists(e.value.errors, errors.InvalidRootNodeError, msg)


def test_invalid_base_uri_wrong_type():
    raml = load_raml("invalid-base-uri-params.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("Validation errors were found.",)
    msg1 = ("baseUriParameter 'domainName' must be a string",)
    assert e.value.args == msg
    assert _error_exists(e.value.errors, errors.InvalidRootNodeError, msg1)


def test_invalid_base_uri_optional():
    raml = load_raml("optional-base-uri-params.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("baseUriParameter 'domainName' must be required",)
    assert _error_exists(e.value.errors, errors.InvalidRootNodeError, msg)


def test_invalid_uri_params_version():
    raml = load_raml("version-in-uri-params.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("'version' can only be defined in baseUriParameters.",)
    assert _error_exists(e.value.errors, errors.InvalidRootNodeError, msg)


def test_invalid_no_title():
    raml = load_raml("no-title.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ('RAML File does not define an API title.',)
    assert _error_exists(e.value.errors, errors.InvalidRootNodeError, msg)


def test_invalid_docs_not_list():
    raml = load_raml("docs-not-list.raml")
    config = load_config("valid-config.ini")
    with pytest.raises(AssertionError) as e:
        validate(raml, config)
    assert ("Error parsing documentation",) == e.value.args


def test_invalid_docs_no_title():
    raml = load_raml("docs-no-title.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("API Documentation requires a title.",)
    assert _error_exists(e.value.errors, errors.InvalidRootNodeError, msg)


def test_invalid_docs_no_content():
    raml = load_raml("docs-no-content.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("API Documentation requires content defined.",)
    assert _error_exists(e.value.errors, errors.InvalidRootNodeError, msg)


def test_assigned_undefined_resource_type():
    raml = load_raml("undefined-resource-type-str.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("Resource Type 'undefined' is assigned to '/foo' but is not "
           "defined in the root of the API.",)
    assert _error_exists(e.value.errors, errors.InvalidResourceNodeError, msg)


def test_no_resources_defined():
    raml = load_raml("no-resources.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("API does not define any resources.",)
    assert _error_exists(e.value.errors, errors.InvalidRootNodeError, msg)


def test_invalid_media_type():
    raml = load_raml("invalid-media-type.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("Unsupported MIME Media Type: 'awesome/sauce'.",)
    assert _error_exists(e.value.errors, errors.InvalidRootNodeError, msg)


# TODO: move assert from parser to validate
def test_invalid_trait_obj():
    raml = load_raml("trait-unsupported-obj.raml")
    config = load_config("valid-config.ini")
    with pytest.raises(AssertionError) as e:
        validate(raml, config)
    msg = ("Error parsing trait",)
    assert msg == e.value.args


def test_traits_undefined():
    raml = load_raml("trait-undefined.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("Trait 'undefined' is assigned to '/users/{user_id}/playlists' "
           "but is not defined in the root of the API.",)
    assert _error_exists(e.value.errors, errors.InvalidResourceNodeError, msg)


def test_no_traits_defined():
    raml = load_raml("no-traits-defined.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("Trying to assign traits that are not defined"
           "in the root of the API.",)
    assert _error_exists(e.value.errors, errors.InvalidResourceNodeError, msg)


# TODO: move assert from parser to validate
def test_unsupported_trait_type_str():
    raml = load_raml("trait-unsupported-type-str.raml")
    config = load_config("valid-config.ini")
    with pytest.raises(AssertionError) as e:
        validate(raml, config)
    msg = ("Error parsing trait",)
    assert msg == e.value.args


def test_unsupported_trait_type_array_ints():
    raml = load_raml("trait-unsupported-type-array-ints.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("'12' needs to be a string referring to a trait, or a dictionary "
           "mapping parameter values to a trait",)
    assert _error_exists(e.value.errors, errors.InvalidResourceNodeError, msg)


def test_too_many_assigned_resource_types():
    raml = load_raml("too-many-assigned-res-types.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("Too many resource types applied to '/foobar'.",)
    assert _error_exists(e.value.errors, errors.InvalidResourceNodeError, msg)


#####
# Parameter Validators
#####
def test_invalid_request_header_param():
    raml = load_raml("invalid-parameter-type-header.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("'invalidType' is not a valid primative parameter type",)
    assert _error_exists(e.value.errors, errors.InvalidParameterError, msg)


def test_invalid_body_mime_type():
    raml = load_raml("invalid-body-mime-type.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("Unsupported MIME Media Type: 'invalid/mediatype'.",)
    assert _error_exists(e.value.errors, errors.InvalidParameterError, msg)


def test_invalid_body_schema():
    raml = load_raml("invalid-body-form-schema.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("Validation errors were found.",)
    msg1 = ("Body must define formParameters, not schema/example.",)
    msg2 = ("Body with mime_type 'application/x-www-form-urlencoded' "
            "requires formParameters.",)
    assert msg == e.value.args
    assert _error_exists(e.value.errors, errors.InvalidParameterError, msg1)
    assert _error_exists(e.value.errors, errors.InvalidParameterError, msg2)


def test_invalid_body_example():
    raml = load_raml("invalid-body-form-example.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)

    msg = ("Validation errors were found.",)
    msg1 = ("Body must define formParameters, not schema/example.",)
    msg2 = ("Body with mime_type 'application/x-www-form-urlencoded' "
            "requires formParameters.",)
    assert msg == e.value.args
    assert _error_exists(e.value.errors, errors.InvalidParameterError, msg1)
    assert _error_exists(e.value.errors, errors.InvalidParameterError, msg2)


def test_invalid_body_no_form_params():
    raml = load_raml("invalid-body-no-form-params.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("Body with mime_type 'application/x-www-form-urlencoded' requires "
           "formParameters.",)
    assert _error_exists(e.value.errors, errors.InvalidParameterError, msg)


def test_invalid_response_code_str():
    raml = load_raml("invalid-response-code-str.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = (
        "Response code 'foo' must be an integer representing an HTTP code.",
    )
    assert _error_exists(e.value.errors, errors.InvalidParameterError, msg)


def test_invalid_response_code():
    raml = load_raml("invalid-response-code.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("'299' not a valid HTTP response code.",)
    assert _error_exists(e.value.errors, errors.InvalidParameterError, msg)


#####
# Primative Validators
#####

def test_invalid_integer_number_type():
    raml = load_raml("invalid-integer-number-type.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("invalidParamType must be either a number or integer to have "
           "minimum attribute set, not 'string'.",)
    assert _error_exists(e.value.errors, errors.InvalidParameterError, msg)


def test_invalid_string_type():
    raml = load_raml("invalid-string-type.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("invalidParamType must be a string type to have min_length "
           "attribute set, not 'integer'.",)
    assert _error_exists(e.value.errors, errors.InvalidParameterError, msg)


#####
# ResourceType, Trait, and Security Scheme validators
#####

def test_empty_mapping_res_type():
    raml = load_raml("empty-mapping-resource-type.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)

    msg = ("The resourceType 'emptyType' requires definition.",)
    assert _error_exists(e.value.errors, errors.InvalidResourceNodeError, msg)


def test_empty_mapping_trait():
    raml = load_raml("empty-mapping-trait.raml")
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)

    msg = ("The trait 'emptyTrait' requires definition.",)
    assert _error_exists(e.value.errors, errors.InvalidResourceNodeError, msg)


def test_empty_mapping_sec_scheme_settings():
    _raml = "empty-mapping-security-scheme-settings.raml"
    raml = load_raml(_raml)
    config = load_config("valid-config.ini")
    with raises as e:
        validate(raml, config)
    msg = ("'settings' for security scheme 'EmptySettingsScheme' require "
           "definition.",)
    assert _error_exists(e.value.errors, errors.InvalidSecuritySchemeError,
                         msg)



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

__all__ = ["RAMLLoader"]

try:
    from collections import OrderedDict
except ImportError:  # pragma: no cover
    from ordereddict import OrderedDict

import os

import jsonref
import yaml

from .errors import LoadRAMLError


class RAMLLoader(object):
    """
    Extends YAML loader to load RAML files with ``!include`` tags.
    """
    def _yaml_include(self, loader, node):
        """
        Adds the ability to follow ``!include`` directives within
        RAML Files.
        """
        # Get the path out of the yaml file
        file_name = os.path.join(os.path.dirname(loader.name), node.value)
        file_ext = os.path.splitext(file_name)[1]
        parsable_ext = [".yaml", ".yml", ".raml", ".json"]

        if file_ext not in parsable_ext:
            with open(file_name) as inputfile:
                return inputfile.read()

        if file_ext == ".json":
            return self._parse_json(file_name, os.path.dirname(file_name))

        with open(file_name) as inputfile:
            return yaml.load(inputfile, self._ordered_loader)

    def _parse_json(self, jsonfile, base_path):
        """
        Parses JSON as well as resolves any `$ref`s, including references to
        local files and remote (HTTP/S) files.
        """
        base_path = os.path.abspath(base_path)
        if not base_path.endswith("/"):
            base_path = base_path + "/"
        base_path = "file://" + base_path

        with open(jsonfile, "r") as f:
            schema = jsonref.load(f, base_uri=base_path, jsonschema=True)
        return schema

    def _ordered_load(self, stream, loader=yaml.SafeLoader):
        """
        Preserves order set in RAML file.
        """
        class OrderedLoader(loader):
            pass

        def construct_mapping(loader, node):
            loader.flatten_mapping(node)
            return OrderedDict(loader.construct_pairs(node))
        OrderedLoader.add_constructor("!include", self._yaml_include)
        OrderedLoader.add_constructor(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping)

        self._ordered_loader = OrderedLoader

        return yaml.load(stream, OrderedLoader)

    def load(self, raml):
        """
        Loads the desired RAML file and returns data.

        :param raml: Either a string/unicode path to RAML file,
            a file object, or string-representation of RAML.

        :return: Data from RAML file
        :rtype: ``dict``

        """

        try:
            return self._ordered_load(raml, yaml.SafeLoader)
        except yaml.parser.ParserError as e:
            msg = "Error parsing RAML: {0}".format(e)
            raise LoadRAMLError(msg)
        except yaml.constructor.ConstructorError as e:
            msg = "Error parsing RAML: {0}".format(e)
            raise LoadRAMLError(msg)



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

from __future__ import absolute_import, division, print_function


#####
# Validation Exceptions
#####


class InvalidRAMLError(Exception):
    def __init__(self, errors):
        super(InvalidRAMLError, self).__init__(
            "Validation errors were found.")
        self.errors = errors

    def __str__(self):
        output = "\n"
        for e in self.errors:
            output += "\t{0}: {1}\n".format(e.__class__.__name__, e)
        # Strip last newline
        return output[:-1]


class BaseRAMLError(Exception):
    pass


class InvalidRootNodeError(BaseRAMLError):
    pass


class InvalidResourceNodeError(BaseRAMLError):
    pass


class InvalidParameterError(BaseRAMLError):
    def __init__(self, message, parameter):
        super(InvalidParameterError, self).__init__(message)
        self.parameter = parameter


class InvalidSecuritySchemeError(BaseRAMLError):
    pass


#####
# Loader Exceptions
#####

class LoadRAMLError(Exception):
    pass


#####
# Update MIME Media Type Exception
#####

class MediaTypeError(Exception):
    pass



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB
from __future__ import absolute_import, division, print_function

import os

import pytest

from ramlfications import tree, parser
from ramlfications.config import setup_config
from ramlfications._helpers import load_file

from .base import EXAMPLES
from .data.fixtures import tree_fixtures


@pytest.fixture(scope="session")
def api():
    raml_str = os.path.join(EXAMPLES, "simple-tree.raml")
    loaded_raml = load_file(raml_str)
    config_file = os.path.join(EXAMPLES + "test-config.ini")
    config = setup_config(config_file)
    return parser.parse_raml(loaded_raml, config)


def print_tree(root, color, verbosity):
    resources = tree._get_tree(root)
    return tree._print_tree(root, resources, color, verbosity)


def test_print_tree_no_color(api, capsys):
    expected_result = tree_fixtures.tree_no_color
    print_tree(api, None, 0)

    out, err = capsys.readouterr()

    assert out == expected_result


def test_print_tree_light(api, capsys):
    expected_result = tree_fixtures.tree_light
    print_tree(api, "light", 0)

    out, err = capsys.readouterr()

    assert out == expected_result


def test_print_tree_dark(api, capsys):
    expected_result = tree_fixtures.tree_dark
    print_tree(api, "dark", 0)

    out, err = capsys.readouterr()

    assert out == expected_result


@pytest.mark.skipif(1 == 1, reason="HALP! I AM COLORBLIND")
def test_print_tree_light_v(api, capsys):
    expected_result = tree_fixtures.tree_light_v
    print_tree(api, "light", 1)

    out, err = capsys.readouterr()
    print(out)
    print(expected_result)
    assert out == expected_result


@pytest.mark.skipif(1 == 1, reason="HALP! I AM COLORBLIND")
def test_print_tree_light_vv(api, capsys):
    expected_result = tree_fixtures.tree_light_vv
    print_tree(api, "light", 2)

    out, err = capsys.readouterr()

    assert out == expected_result


@pytest.mark.skipif(1 == 1, reason="HALP! I AM COLORBLIND")
def test_print_tree_light_vvv(api, capsys):
    expected_result = tree_fixtures.tree_light_vvv
    print_tree(api, "light", 3)

    out, err = capsys.readouterr()

    assert out == expected_result



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

from __future__ import absolute_import, division, print_function

import re

from six import iterkeys

from ._decorators import collecterrors

from .errors import *  # NOQA


#####
# RAMLRoot validators
#####

@collecterrors
def root_version(inst, attr, value):
    """Require an API Version (e.g. api.foo.com/v1)."""
    base_uri = inst.raml_obj.get("baseUri")
    if base_uri and "{version}" in base_uri and not value:
        msg = ("RAML File's baseUri includes {version} parameter but no "
               "version is defined.")
        raise InvalidRootNodeError(msg)


@collecterrors
def root_base_uri(inst, attr, value):
    """Require a Base URI."""
    if not value:
        msg = "RAML File does not define the baseUri."
        raise InvalidRootNodeError(msg)


@collecterrors
def root_base_uri_params(inst, attr, value):
    """
    Require that Base URI parameters have a ``default`` parameter set.
    """
    if value:
        for v in value:
            if v.type != "string":
                msg = ("baseUriParameter '{0}' must be a "
                       "string".format(v.name))
                raise InvalidRootNodeError(msg)
            if v.required is False:
                msg = ("baseUriParameter '{0}' must be "
                       "required".format(v.name))
                raise InvalidRootNodeError(msg)


@collecterrors
def root_uri_params(inst, attr, value):
    """
    Assert that where is no ``version`` parameter in the regular URI parameters
    """
    if value:
        for v in value:
            if v.name == "version":
                msg = "'version' can only be defined in baseUriParameters."
                raise InvalidRootNodeError(msg)


@collecterrors
def root_protocols(inst, attr, value):
    """
    Only support HTTP/S plus what is defined in user-config
    """
    if value:
        for p in value:
            if p.upper() not in inst.config.get("protocols"):
                msg = ("'{0}' not a valid protocol for a RAML-defined "
                       "API.".format(p))
                raise InvalidRootNodeError(msg)


@collecterrors
def root_title(inst, attr, value):
    """
    Require a title for the defined API.
    """
    if not value:
        msg = "RAML File does not define an API title."
        raise InvalidRootNodeError(msg)


@collecterrors
def root_docs(inst, attr, value):
    """
    Assert that if there is ``documentation`` defined in the root of the
    RAML file, that it contains a ``title`` and ``content``.
    """
    if value:
        for d in value:
            if d.title.raw is None:
                msg = "API Documentation requires a title."
                raise InvalidRootNodeError(msg)
            if d.content.raw is None:
                msg = "API Documentation requires content defined."
                raise InvalidRootNodeError(msg)


# TODO: finish
@collecterrors
def root_schemas(inst, attr, value):
    pass


@collecterrors
def root_media_type(inst, attr, value):
    """
    Only support media types based on config and regex
    """
    if value:
        match = validate_mime_type(value)
        if value not in inst.config.get("media_types") and not match:
            msg = "Unsupported MIME Media Type: '{0}'.".format(value)
            raise InvalidRootNodeError(msg)


@collecterrors
def root_resources(inst, attr, value):
    if not value:
        msg = "API does not define any resources."
        raise InvalidRootNodeError(msg)


@collecterrors
def root_secured_by(inst, attr, value):
    pass


#####
# Security Scheme Validators
#####
@collecterrors
def defined_sec_scheme_settings(inst, attr, value):
    """Assert that ``settings`` are defined/not an empty map."""
    if not value:
        msg = ("'settings' for security scheme '{0}' require "
               "definition.".format(inst.name))
        raise InvalidSecuritySchemeError(msg)


#####
# ResourceType validators
#####
@collecterrors
def defined_resource_type(inst, attr, value):
    """
    Assert that a resource type is defined (e.g. not an empty map) or is
    inherited from a defined resource type.
    """
    if not inst.type:
        if not value:
            msg = ("The resourceType '{0}' requires "
                   "definition.".format(inst.name))
            raise InvalidResourceNodeError(msg)


#####
# Trait validators
#####
@collecterrors
def defined_trait(inst, attr, value):
    """Assert that a trait is defined (e.g. not an empty map)."""
    if not value:
        msg = "The trait '{0}' requires definition.".format(inst.name)
        raise InvalidResourceNodeError(msg)


#####
# Shared Validators for Resource & Resource Type Node
#####
@collecterrors
def assigned_traits(inst, attr, value):
    """
    Assert assigned traits are defined in the RAML Root and correctly
    represented in the RAML.
    """
    if value:
        traits = inst.root.raw.get("traits", {})
        if not traits:
            msg = ("Trying to assign traits that are not defined"
                   "in the root of the API.")
            raise InvalidResourceNodeError(msg)
        trait_names = [list(iterkeys(i))[0] for i in traits]
        if isinstance(value, list):
            for v in value:
                if isinstance(v, dict):
                    if list(iterkeys(v))[0] not in trait_names:  # NOCOV
                        msg = (
                            "Trait '{0}' is assigned to '{1}' but is not def"
                            "ined in the root of the API.".format(v, inst.path)
                        )
                        raise InvalidResourceNodeError(msg)
                    if not isinstance(list(iterkeys(v))[0], str):  # NOCOV
                        msg = ("'{0}' needs to be a string referring to a "
                               "trait, or a dictionary mapping parameter "
                               "values to a trait".format(v))
                        raise InvalidResourceNodeError(msg)
                elif isinstance(v, str):
                    if v not in trait_names:
                        msg = (
                            "Trait '{0}' is assigned to '{1}' but is not "
                            "defined in the root of the API.".format(v,
                                                                     inst.path)
                        )
                        raise InvalidResourceNodeError(msg)
                else:
                    msg = ("'{0}' needs to be a string referring to a "
                           "trait, or a dictionary mapping parameter "
                           "values to a trait".format(v))
                    raise InvalidResourceNodeError(msg)


@collecterrors
def assigned_res_type(inst, attr, value):
    """
    Assert only one (or none) assigned resource type is defined in the RAML
    Root and correctly represented in the RAML.
    """
    if value:
        if isinstance(value, tuple([dict, list])) and len(value) > 1:
            msg = "Too many resource types applied to '{0}'.".format(
                inst.display_name
            )
            raise InvalidResourceNodeError(msg)

        res_types = inst.root.raw.get("resourceTypes", {})
        res_type_names = [list(iterkeys(i))[0] for i in res_types]
        if isinstance(value, list):
            item = value[0]  # NOCOV
        elif isinstance(value, dict):
            item = list(iterkeys(value))[0]  # NOCOV
        else:
            item = value
        if item not in res_type_names:
            msg = ("Resource Type '{0}' is assigned to '{1}' but is not "
                   "defined in the root of the API.".format(value,
                                                            inst.display_name))
            raise InvalidResourceNodeError(msg)


#####
# Parameter Validators
#####
@collecterrors
def header_type(inst, attr, value):
    """Supported header type"""
    if value and value not in inst.config.get("prim_types"):
        msg = "'{0}' is not a valid primative parameter type".format(value)
        raise InvalidParameterError(msg, "header")


@collecterrors
def body_mime_type(inst, attr, value):
    """Supported MIME media type for request/response"""
    if value:
        match = validate_mime_type(value)
        if value not in inst.config.get("media_types") and not match:
            msg = "Unsupported MIME Media Type: '{0}'.".format(value)
            raise InvalidParameterError(msg, "body")


@collecterrors
def body_schema(inst, attr, value):
    """
    Assert no ``schema`` is defined if body as a form-related MIME media type
    """
    form_types = ["multipart/form-data", "application/x-www-form-urlencoded"]
    if inst.mime_type in form_types and value:
        msg = "Body must define formParameters, not schema/example."
        raise InvalidParameterError(msg, "body")


@collecterrors
def body_example(inst, attr, value):
    """
    Assert no ``example`` is defined if body as a form-related MIME media type
    """
    form_types = ["multipart/form-data", "application/x-www-form-urlencoded"]
    if inst.mime_type in form_types and value:
        msg = "Body must define formParameters, not schema/example."
        raise InvalidParameterError(msg, "body")


@collecterrors
def body_form(inst, attr, value):
    """
    Assert ``formParameters`` are defined if body has a form-related
    MIME type.
    """
    form_types = ["multipart/form-data", "application/x-www-form-urlencoded"]
    if inst.mime_type in form_types and not value:
        msg = "Body with mime_type '{0}' requires formParameters.".format(
            inst.mime_type)
        raise InvalidParameterError(msg, "body")


@collecterrors
def response_code(inst, attr, value):
    """
    Assert a valid response code.
    """
    if not isinstance(value, int):
        msg = ("Response code '{0}' must be an integer representing an "
               "HTTP code.".format(value))
        raise InvalidParameterError(msg, "response")
    if value not in inst.config.get("resp_codes"):
        msg = "'{0}' not a valid HTTP response code.".format(value)
        raise InvalidParameterError(msg, "response")


#####
# Primative Validators
#####
@collecterrors
def integer_number_type_parameter(inst, attr, value):
    """
    Assert correct parameter attributes for ``integer`` or ``number``
    primative parameter types.
    """
    if value is not None:
        param_types = ["integer", "number"]
        if inst.type not in param_types:
            msg = ("{0} must be either a number or integer to have {1} "
                   "attribute set, not '{2}'.".format(inst.name, attr.name,
                                                      inst.type))
            raise InvalidParameterError(msg, "BaseParameter")


@collecterrors
def string_type_parameter(inst, attr, value):
    """
    Assert correct parameter attributes for ``string`` primative parameter
    types.
    """
    if value:
        if inst.type != "string":
            msg = ("{0} must be a string type to have {1} "
                   "attribute set, not '{2}'.".format(inst.name, attr.name,
                                                      inst.type))
            raise InvalidParameterError(msg, "BaseParameter")


#####
# Util/common functions
#####


def validate_mime_type(value):
    """
    Assert a valid MIME media type for request/response body.
    """
    regex_str = re.compile(r"application\/[A-Za-z.-0-1]*?(json|xml)")
    match = re.search(regex_str, value)
    return match



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

import os

from click.testing import CliRunner
import pytest

from ramlfications import __main__ as main

from .base import EXAMPLES, VALIDATE


@pytest.fixture
def runner():
    return CliRunner()


def check_result(exp_code, exp_msg, result):
    assert result.exit_code == exp_code
    if exp_msg:
        assert result.output == exp_msg


def test_validate(runner):
    """
    Successfully validate RAML file via CLI.
    """
    raml_file = os.path.join(EXAMPLES, "complete-valid-example.raml")
    exp_code = 0
    exp_msg = "Success! Valid RAML file: {0}\n".format(raml_file)
    result = runner.invoke(main.validate, [raml_file])
    check_result(exp_code, exp_msg, result)


def test_validate_fail(runner):
    """
    Raise error for invalid RAML file via CLI when validating.
    """
    raml_file = os.path.join(VALIDATE, "no-base-uri-no-title.raml")
    exp_code = 1
    exp_msg_1 = "Error validating file {0}: \n".format(raml_file)
    exp_msg_2 = 'RAML File does not define an API title.'
    exp_msg_3 = 'RAML File does not define the baseUri.'

    result = runner.invoke(main.validate, [raml_file])

    assert result.exit_code == exp_code
    assert exp_msg_1 in result.output
    assert exp_msg_2 in result.output
    assert exp_msg_3 in result.output


def test_tree(runner):
    """
    Successfully print out tree of RAML file via CLI.
    """
    raml_file = os.path.join(EXAMPLES, "complete-valid-example.raml")
    config_file = os.path.join(EXAMPLES, "test-config.ini")
    exp_code = 0
    exp_msg = None
    result = runner.invoke(main.tree, [raml_file, "--color=light",
                           "--config={0}".format(config_file)])
    check_result(exp_code, exp_msg, result)


def test_tree_invalid(runner):
    """
    Raise error for invalid RAML file via CLI when printing the tree.
    """
    raml_file = os.path.join(VALIDATE, "no-title.raml")
    config_file = os.path.join(VALIDATE, "valid-config.ini")
    exp_code = 1
    exp_msg = '"{0}" is not a valid RAML file: \n\t{1}: {2}\n'.format(
        raml_file, 'InvalidRootNodeError',
        'RAML File does not define an API title.')
    result = runner.invoke(main.tree, [raml_file, "--color=light",
                           "--config={0}".format(config_file)])
    check_result(exp_code, exp_msg, result)


def test_update(runner, mocker):
    """
    Successfully update supported mime types
    """
    json_file = "ramlfications/data/supported_mime_types.json"
    parent = os.path.dirname(os.path.pardir)
    json_path = os.path.join(parent, json_file)

    start_mtime = os.path.getmtime(json_path)

    from ramlfications import utils
    mocker.patch("ramlfications.utils.update_mime_types")
    mocker.patch("ramlfications.utils._save_updated_mime_types")

    runner.invoke(main.update)

    utils.update_mime_types.assert_called_once()
    utils._save_updated_mime_types.assert_called_once()

    end_mtime = os.path.getmtime(json_path)

    # sanity check that data was not written to file
    assert start_mtime == end_mtime



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

from __future__ import absolute_import, division, print_function
import re

import attr
from six import iteritems, iterkeys, itervalues


from .config import MEDIA_TYPES
from .errors import InvalidRAMLError
from .parameters import (
    Documentation, Header, Body, Response, URIParameter, QueryParameter,
    FormParameter, SecurityScheme
)
from .parser_utils import (
    security_schemes
)
from .raml import RootNode, ResourceNode, ResourceTypeNode, TraitNode
from .utils import (
    load_schema, _resource_type_lookup,
    _get_resource_type, _get_trait, _get_attribute,
    _get_inherited_attribute, _remove_duplicates, _create_uri_params,
    _get, _create_base_param_obj, _get_res_type_attribute,
    _get_inherited_type_params, _get_inherited_item, _get_attribute_dict,
    get_inherited, set_param_object, set_params, _get_data_union,
    _preserve_uri_order
)


__all__ = ["parse_raml"]


def parse_raml(loaded_raml, config):
    """
    Parse loaded RAML file into RAML/Python objects.

    :param RAMLDict loaded_raml: OrderedDict of loaded RAML file
    :returns: :py:class:`.raml.RootNode` object.
    :raises: :py:class:`.errors.InvalidRAMLError` when RAML file is invalid
    """

    validate = str(_get(config, "validate")).lower() == 'true'

    # Postpone validating the root node until the end; otherwise,
    # we end up with duplicate validation exceptions.
    attr.set_run_validators(False)
    root = create_root(loaded_raml, config)
    attr.set_run_validators(validate)

    root.security_schemes = create_sec_schemes(root.raml_obj, root)
    root.traits = create_traits(root.raml_obj, root)
    root.resource_types = create_resource_types(root.raml_obj, root)
    root.resources = create_resources(root.raml_obj, [], root,
                                      parent=None)

    if validate:
        attr.validate(root)  # need to validate again for root node

        if root.errors:
            raise InvalidRAMLError(root.errors)

    return root


def create_root(raml, config):
    """
    Creates a Root Node based off of the RAML's root section.

    :param RAMLDict raml: loaded RAML file
    :returns: :py:class:`.raml.RootNode` object with API root attributes set
    """

    errors = []

    def protocols():
        explicit_protos = _get(raml, "protocols")
        implicit_protos = re.findall(r"(https|http)", base_uri())
        implicit_protos = [p.upper() for p in implicit_protos]

        return explicit_protos or implicit_protos or None

    def base_uri():
        base_uri = _get(raml, "baseUri", "")
        if "{version}" in base_uri:
            base_uri = base_uri.replace("{version}",
                                        str(_get(raml, "version")))
        return base_uri

    def base_uri_params():
        data = _get(raml, "baseUriParameters", {})
        params = _create_base_param_obj(data, URIParameter, config, errors)
        uri = _get(raml, "baseUri", "")
        declared = _get(raml, "uriParameters", {})
        declared = list(iterkeys(declared))
        return _preserve_uri_order(uri, params, config, errors, declared)

    def uri_params():
        data = _get(raml, "uriParameters", {})
        params = _create_base_param_obj(data, URIParameter, config, errors)
        uri = _get(raml, "baseUri", "")

        declared = []
        base = base_uri_params()
        if base:
            declared = [p.name for p in base]
        return _preserve_uri_order(uri, params, config, errors, declared)

    def docs():
        d = _get(raml, "documentation", [])
        assert isinstance(d, list), "Error parsing documentation"
        docs = [Documentation(_get(i, "title"), _get(i, "content")) for i in d]
        return docs or None

    def schemas():
        _schemas = _get(raml, "schemas")
        if not _schemas:
            return None
        schemas = []
        for schema in _schemas:
            value = load_schema(list(itervalues(schema))[0])
            schemas.append({list(iterkeys(schema))[0]: value})
        return schemas or None

    return RootNode(
        raml_obj=raml,
        raw=raml,
        title=_get(raml, "title"),
        version=_get(raml, "version"),
        protocols=protocols(),
        base_uri=base_uri(),
        base_uri_params=base_uri_params(),
        uri_params=uri_params(),
        media_type=_get(raml, "mediaType"),
        documentation=docs(),
        schemas=schemas(),
        config=config,
        secured_by=_get(raml, "securedBy"),
        errors=errors
    )


def create_sec_schemes(raml_data, root):
    """
    Parse security schemes into ``SecurityScheme`` objects

    :param dict raml_data: Raw RAML data
    :param RootNode root: Root Node
    :returns: list of :py:class:`.parameters.SecurityScheme` objects
    """
    def _map_object_types(item):
        return {
            "headers": headers,
            "body": body,
            "responses": responses,
            "queryParameters": query_params,
            "uriParameters": uri_params,
            "formParameters": form_params,
            "usage": usage,
            "mediaType": media_type,
            "protocols": protocols,
            "documentation": documentation,
        }[item]

    def headers(header_data):
        _headers = []
        header_data = _get(header_data, "headers", {})
        for k, v in list(iteritems(header_data)):
            h = _create_base_param_obj({k: v},
                                       Header,
                                       root.config,
                                       root.errors)
            _headers.extend(h)
        return _headers

    def body(body_data):
        body_data = _get(body_data, "body", {})
        _body = []
        for k, v in list(iteritems(body_data)):
            body = Body(
                mime_type=k,
                raw=v,
                schema=load_schema(_get(v, "schema")),
                example=load_schema(_get(v, "example")),
                form_params=_get(v, "formParameters"),
                config=root.config,
                errors=root.errors
            )
            _body.append(body)
        return _body

    def responses(resp_data):
        _resps = []
        resp_data = _get(resp_data, "responses", {})
        for k, v in list(iteritems(resp_data)):
            response = Response(
                code=k,
                raw=v,
                desc=_get(v, "description"),
                headers=headers(_get(v, "headers", {})),
                body=body(_get(v, "body", {})),
                config=root.config,
                errors=root.errors
            )
            _resps.append(response)
        return sorted(_resps, key=lambda x: x.code)

    def query_params(param_data):
        param_data = _get(param_data, "queryParameters", {})
        _params = []
        for k, v in list(iteritems(param_data)):
            p = _create_base_param_obj({k: v},
                                       QueryParameter,
                                       root.config,
                                       root.errors)
            _params.extend(p)
        return _params

    def uri_params(param_data):
        param_data = _get(param_data, "uriParameters")
        _params = []
        for k, v in list(iteritems(param_data)):
            p = _create_base_param_obj({k: v},
                                       URIParameter,
                                       root.config,
                                       root.errors)
            _params.extend(p)
        return _params

    def form_params(param_data):
        param_data = _get(param_data, "formParameters", {})
        _params = []
        for k, v in list(iteritems(param_data)):
            p = _create_base_param_obj({k: v},
                                       FormParameter,
                                       root.config,
                                       root.errors)
            _params.extend(p)
        return _params

    def usage(desc_by_data):
        return _get(desc_by_data, "usage")

    def media_type(desc_by_data):
        return _get(desc_by_data, "mediaType")

    def protocols(desc_by_data):
        return _get(desc_by_data, "protocols")

    def documentation(desc_by_data):
        d = _get(desc_by_data, "documentation", [])
        assert isinstance(d, list), "Error parsing documentation"
        docs = [Documentation(_get(i, "title"), _get(i, "content")) for i in d]
        return docs or None

    def set_property(node, obj, node_data):
        func = _map_object_types(obj)
        item_objs = func({obj: node_data})
        setattr(node, func.__name__, item_objs)

    def initial_wrap(key, data):
        return SecurityScheme(
            name=key,
            raw=data,
            type=_get(data, "type"),
            described_by=_get(data, "describedBy", {}),
            desc=_get(data, "description"),
            settings=_get(data, "settings"),
            config=root.config,
            errors=root.errors
        )

    def final_wrap(node):
        for obj, node_data in list(iteritems(node.described_by)):
            set_property(node, obj, node_data)
        return node

    schemes = _get(raml_data, "securitySchemes", [])
    scheme_objs = []
    for s in schemes:
        name = list(iterkeys(s))[0]
        data = list(itervalues(s))[0]
        node = initial_wrap(name, data)
        node = final_wrap(node)
        scheme_objs.append(node)
    return scheme_objs or None


def create_traits(raml_data, root):
    """
    Parse traits into ``Trait`` objects.

    :param dict raml_data: Raw RAML data
    :param RootNode root: Root Node
    :returns: list of :py:class:`.raml.TraitNode` objects
    """
    def description():
        return _get(data, "description")

    def protocols():
        return _get(data, "protocols")

    def query_params():
        return set_param_object(data, "queryParameters", root)

    def uri_params():
        return set_param_object(data, "uriParameters", root)

    def form_params():
        return set_param_object(data, "formParameters", root)

    def base_uri_params():
        return set_param_object(data, "baseUriParameters", root)

    def headers(data):
        return set_param_object(data, "headers", root)

    def body(data):
        body = _get(data, "body", {})
        body_objects = []
        for key, value in list(iteritems(body)):
            body = Body(
                mime_type=key,
                raw=value,
                schema=load_schema(_get(value, "schema")),
                example=load_schema(_get(value, "example")),
                form_params=_get(value, "formParameters"),
                config=root.config,
                errors=root.errors
            )
            body_objects.append(body)
        return body_objects or None

    def responses():
        response_objects = []
        for key, value in list(iteritems(_get(data, "responses", {}))):
            response = Response(
                code=key,
                raw=value,
                desc=_get(value, "description"),
                headers=headers(value),
                body=body(value),
                config=root.config,
                errors=root.errors
            )
            response_objects.append(response)
        return sorted(response_objects, key=lambda x: x.code) or None

    def wrap(key, data):
        return TraitNode(
            name=key,
            raw=data,
            root=root,
            query_params=query_params(),
            uri_params=uri_params(),
            form_params=form_params(),
            base_uri_params=base_uri_params(),
            headers=headers(data),
            body=body(data),
            responses=responses(),
            desc=description(),
            media_type=_get(data, "mediaType"),
            usage=_get(data, "usage"),
            protocols=protocols(),
            errors=root.errors
        )

    traits = _get(raml_data, "traits", [])
    trait_objects = []
    for trait in traits:
        name = list(iterkeys(trait))[0]
        data = list(itervalues(trait))[0]
        trait_objects.append(wrap(name, data))
    return trait_objects or None


def create_resource_types(raml_data, root):
    """
    Parse resourceTypes into ``ResourceTypeNode`` objects.

    :param dict raml_data: Raw RAML data
    :param RootNode root: Root Node
    :returns: list of :py:class:`.raml.ResourceTypeNode` objects
    """
    # TODO: move this outside somewhere - config?
    accepted_methods = _get(root.config, "http_optional")

    #####
    # Set ResourceTypeNode attributes
    #####

    def headers(data):
        _headers = _get(data, "headers", {})
        if _get(v, "type"):
            _headers = _get_inherited_item(_headers, "headers",
                                           resource_types,
                                           meth, v)

        header_objs = _create_base_param_obj(_headers,
                                             Header,
                                             root.config,
                                             root.errors)
        if header_objs:
            for h in header_objs:
                h.method = method(meth)

        return header_objs

    def body(data):
        _body = _get(data, "body", default={})
        if _get(v, "type"):
            _body = _get_inherited_item(_body, "body", resource_types,
                                        meth, v)

        body_objects = []
        for key, value in list(iteritems(_body)):
            body = Body(
                mime_type=key,
                raw=value,
                schema=load_schema(_get(value, "schema")),
                example=load_schema(_get(value, "example")),
                form_params=_get(value, "formParameters"),
                config=root.config,
                errors=root.errors
            )
            body_objects.append(body)
        return body_objects or None

    def responses(data):
        response_objects = []
        _responses = _get(data, "responses", {})
        if _get(v, "type"):
            _responses = _get_inherited_item(_responses, "responses",
                                             resource_types, meth, v)

        for key, value in list(iteritems(_responses)):
            _headers = _get(_get(data, "responses", {}), key, {})
            _headers = _get(_headers, "headers", {})
            header_objs = _create_base_param_obj(_headers, Header,
                                                 root.config, root.errors)
            if header_objs:
                for h in header_objs:
                    h.method = method(meth)
            response = Response(
                code=key,
                raw={key: value},
                desc=_get(value, "description"),
                headers=header_objs,
                body=body(value),
                config=root.config,
                method=method(meth),
                errors=root.errors
            )
            response_objects.append(response)
        if response_objects:
            return sorted(response_objects, key=lambda x: x.code)
        return None

    def uri_params(data):
        uri_params = _get_attribute_dict(data, "uriParameters", v)

        if _get(v, "type"):
            uri_params = _get_inherited_type_params(v, "uriParameters",
                                                    uri_params, resource_types)
        return _create_base_param_obj(uri_params,
                                      URIParameter,
                                      root.config,
                                      root.errors)

    def base_uri_params(data):
        uri_params = _get_attribute_dict(data, "baseUriParameters", v)

        return _create_base_param_obj(uri_params,
                                      URIParameter,
                                      root.config,
                                      root.errors)

    def query_params(data):
        query_params = _get_attribute_dict(data, "queryParameters", v)

        if _get(v, "type"):
            query_params = _get_inherited_type_params(v, "queryParameters",
                                                      query_params,
                                                      resource_types)

        return _create_base_param_obj(query_params,
                                      QueryParameter,
                                      root.config,
                                      root.errors)

    def form_params(data):
        form_params = _get_attribute_dict(data, "formParameters", v)

        if _get(v, "type"):
            form_params = _get_inherited_type_params(v, "formParameters",
                                                     form_params,
                                                     resource_types)

        return _create_base_param_obj(form_params,
                                      FormParameter,
                                      root.config,
                                      root.errors)

    def description():
        # prefer the resourceType method description
        if meth:
            method_attr = _get(v, meth)
            desc = _get(method_attr, "description")
            return desc or _get(v, "description")
        return _get(v, "description")

    def type_():
        return _get(v, "type")

    def method(meth):
        if not meth:
            return None
        if "?" in meth:
            return meth[:-1]
        return meth

    def optional():
        if meth:
            return "?" in meth

    def protocols(data):
        m, r = _get_res_type_attribute(v, data, "protocols", None)
        return m or r or root.protocols

    def is_(data):
        m, r = _get_res_type_attribute(v, data, "is", default=[])
        return m + r or None

    def traits(data):
        assigned = is_(data)
        if assigned:
            if root.traits:
                trait_objs = []
                for trait in assigned:
                    obj = [t for t in root.traits if t.name == trait]
                    if obj:
                        trait_objs.append(obj[0])
                return trait_objs or None

    def secured_by(data):
        m, r = _get_res_type_attribute(v, data, "securedBy", [])
        return m + r or None

    def security_schemes_(data):
        secured = secured_by(data)
        return security_schemes(secured, root)

    def wrap(key, data, meth, _v):
        return ResourceTypeNode(
            name=key,
            raw=data,
            root=root,
            headers=headers(data),
            body=body(data),
            responses=responses(data),
            uri_params=uri_params(data),
            base_uri_params=base_uri_params(data),
            query_params=query_params(data),
            form_params=form_params(data),
            media_type=_get(v, "mediaType"),
            desc=description(),
            type=type_(),
            method=method(meth),
            usage=_get(v, "usage"),
            optional=optional(),
            is_=is_(data),
            traits=traits(data),
            secured_by=secured_by(data),
            security_schemes=security_schemes_(data),
            display_name=_get(data, "displayName", key),
            protocols=protocols(data),
            errors=root.errors
        )

    resource_types = _get(raml_data, "resourceTypes", [])
    resource_type_objects = []
    child_res_type_objects = []
    child_res_type_names = []

    for res in resource_types:
        for k, v in list(iteritems(res)):
            if isinstance(v, dict):
                if "type" in list(iterkeys(v)):
                    child_res_type_objects.append({k: v})
                    child_res_type_names.append(k)

                else:
                    for meth in list(iterkeys(v)):
                        if meth in accepted_methods:
                            method_data = _get(v, meth, {})
                            resource = wrap(k, method_data, meth, v)
                            resource_type_objects.append(resource)
            else:
                meth = None
                resource = wrap(k, {}, meth, v)
                resource_type_objects.append(resource)

    while child_res_type_objects:
        child = child_res_type_objects.pop()
        name = list(iterkeys(child))[0]
        data = list(itervalues(child))[0]
        parent = data.get("type")
        if parent in child_res_type_names:
            continue
        p_data = [r for r in resource_types if list(iterkeys(r))[0] == parent]
        p_data = p_data[0].get(parent)
        res_data = _get_data_union(data, p_data)

        for meth in list(iterkeys(res_data)):
            if meth in accepted_methods:
                method_data = _get(res_data, meth, {})
                comb_data = dict(list(iteritems(method_data)) +
                                 list(iteritems(res_data)))
                resource = ResourceTypeNode(
                    name=name,
                    raw=res_data,
                    root=root,
                    headers=headers(method_data),
                    body=body(method_data),
                    responses=responses(method_data),
                    uri_params=uri_params(comb_data),
                    base_uri_params=base_uri_params(comb_data),
                    query_params=query_params(method_data),
                    form_params=form_params(method_data),
                    media_type=_get(v, "mediaType"),
                    desc=description(),
                    type=_get(res_data, "type"),
                    method=method(meth),
                    usage=_get(res_data, "usage"),
                    optional=optional(),
                    is_=is_(res_data),
                    traits=traits(res_data),
                    secured_by=secured_by(res_data),
                    security_schemes=security_schemes_(res_data),
                    display_name=_get(method_data, "displayName", name),
                    protocols=protocols(res_data),
                    errors=root.errors
                )
                resource_type_objects.append(resource)

    return resource_type_objects or None


def create_resources(node, resources, root, parent):
    """
    Recursively traverses the RAML file via DFS to find each resource
    endpoint.

    :param dict node: Dictionary of node to traverse
    :param list resources: List of collected ``ResourceNode`` s
    :param RootNode root: The ``RootNode`` of the API
    :param ResourceNode parent: Parent ``ResourceNode`` of current ``node``
    :returns: List of :py:class:`.raml.ResourceNode` objects.
    """
    for k, v in list(iteritems(node)):
        if k.startswith("/"):
            avail = _get(root.config, "http_optional")
            methods = [m for m in avail if m in list(iterkeys(v))]
            if "type" in list(iterkeys(v)):
                assigned = _resource_type_lookup(_get(v, "type"), root)
                if hasattr(assigned, "method"):
                    if not assigned.optional:
                        methods.append(assigned.method)
                        methods = list(set(methods))
            if methods:
                for m in methods:
                    child = create_node(name=k,
                                        raw_data=v,
                                        method=m,
                                        parent=parent,
                                        root=root)
                    resources.append(child)
            # inherit resource type methods
            elif "type" in list(iterkeys(v)):
                if hasattr(assigned, "method"):
                    method = assigned.method
                else:
                    method = None
                child = create_node(name=k,
                                    raw_data=v,
                                    method=method,
                                    parent=parent,
                                    root=root)
                resources.append(child)
            else:
                child = create_node(name=k,
                                    raw_data=v,
                                    method=None,
                                    parent=parent,
                                    root=root)
                resources.append(child)
            resources = create_resources(child.raw, resources, root, child)
    return resources


def create_node(name, raw_data, method, parent, root):
    """
    Create a Resource Node object.

    :param str name: Name of resource node
    :param dict raw_data: Raw RAML data associated with resource node
    :param str method: HTTP method associated with resource node
    :param ResourceNode parent: Parent node object of resource node, if any
    :param RootNode api: API ``RootNode`` that the resource node is attached to
    :returns: :py:class:`.raml.ResourceNode` object
    """
    #####
    # Node attribute functions
    #####
    def path():
        """Set resource's relative URI path."""
        parent_path = ""
        if parent:
            parent_path = parent.path
        return parent_path + name

    def absolute_uri():
        """Set resource's absolute URI path."""
        uri = root.base_uri + path()
        proto = protocols()
        if proto:
            uri = uri.split("://")
            if len(uri) == 2:
                uri = uri[1]
            if root.protocols:
                _proto = list(set(root.protocols) & set(proto))
                # if resource protocols and root protocols share a protocol
                # then use that one
                if _proto:
                    uri = _proto[0].lower() + "://" + uri
                # if no shared protocols, use the first of the resource
                # protocols
                else:
                    uri = proto[0].lower() + "://" + uri
        return uri

    def protocols():
        """Set resource's supported protocols."""
        # trait = _get_trait("protocols", root, is_())
        kwargs = dict(root=root,
                      is_=is_(),
                      type_=type_(),
                      method=method,
                      data=raw_data,
                      parent=parent)
        objects_to_inherit = [
            "traits", "types", "method", "resource", "parent"
        ]
        inherited = get_inherited("protocols", objects_to_inherit, **kwargs)
        trait = inherited["traits"]
        r_type = inherited["types"]
        meth = inherited["method"]
        res = inherited["resource"]
        parent_ = inherited["parent"]
        default = [root.base_uri.split("://")[0].upper()]

        return meth or r_type or trait or res or parent_ or default

    def headers():
        """Set resource's supported headers."""
        headers = _get_attribute("headers", method, raw_data)
        header_objs = _get_inherited_attribute("headers", root, type_(),
                                               method, is_())

        _headers = _create_base_param_obj(headers,
                                          Header,
                                          root.config,
                                          root.errors,
                                          method=method)
        if _headers is None:
            return header_objs or None
        return _remove_duplicates(header_objs, _headers)

    def body():
        """Set resource's supported request/response body."""
        bodies = _get_attribute("body", method, raw_data)
        body_objects = _get_inherited_attribute("body", root, type_(),
                                                method, is_())

        _body_objs = []
        for k, v in list(iteritems(bodies)):
            if v is None:
                continue
            body = Body(
                mime_type=k,
                raw={k: v},
                schema=load_schema(_get(v, "schema")),
                example=load_schema(_get(v, "example")),
                form_params=_get(v, "formParameters"),
                config=root.config,
                errors=root.errors
            )
            _body_objs.append(body)
        if _body_objs == []:
            return body_objects or None
        return _remove_duplicates(body_objects, _body_objs)

    def responses():
        """Set resource's expected responses."""
        def resp_headers(headers):
            """Set response headers."""
            header_objs = []
            for k, v in list(iteritems(headers)):
                header = Header(
                    name=k,
                    display_name=_get(v, "displayName", default=k),
                    method=method,
                    raw=headers,
                    type=_get(v, "type", default="string"),
                    desc=_get(v, "description"),
                    example=_get(v, "example"),
                    default=_get(v, "default"),
                    minimum=_get(v, "minimum"),
                    maximum=_get(v, "maximum"),
                    min_length=_get(v, "minLength"),
                    max_length=_get(v, "maxLength"),
                    enum=_get(v, "enum"),
                    repeat=_get(v, "repeat", default=False),
                    pattern=_get(v, "pattern"),
                    config=root.config,
                    errors=root.errors
                )
                header_objs.append(header)
            return header_objs or None

        def resp_body(body):
            """Set response body."""
            body_list = []
            default_body = {}
            for (key, spec) in body.items():
                if key not in MEDIA_TYPES:
                    # if a root mediaType was defined, the response body
                    # may omit the mime_type definition
                    if key in ('schema', 'example'):
                        default_body[key] = load_schema(spec) if spec else {}
                else:
                    mime_type = key
                    # spec might be '!!null'
                    raw = spec or body
                    _schema = {}
                    _example = {}
                    if spec:
                        _schema_spec = _get(spec, 'schema', '')
                        _example_spec = _get(spec, 'example', '')
                        if _schema_spec:
                            _schema = load_schema(_schema_spec)
                        if _example_spec:
                            _example = load_schema(_example_spec)
                    body_list.append(Body(
                        mime_type=mime_type,
                        raw=raw,
                        schema=_schema,
                        example=_example,
                        form_params=None,
                        config=root.config,
                        errors=root.errors
                    ))
            if default_body:
                body_list.append(Body(
                    mime_type=root.media_type,
                    raw=body,
                    schema=_get(default_body, 'schema'),
                    example=_get(default_body, 'example'),
                    form_params=None,
                    config=root.config,
                    errors=root.errors
                ))

            return body_list or None

        resps = _get_attribute("responses", method, raw_data)
        type_resp = _get_resource_type("responses", root, type_(), method)
        trait_resp = _get_trait("responses", root, is_())
        resp_objs = type_resp + trait_resp
        resp_codes = [r.code for r in resp_objs]
        for k, v in list(iteritems(resps)):
            if k in resp_codes:
                resp = [r for r in resp_objs if r.code == k][0]
                index = resp_objs.index(resp)
                inherit_resp = resp_objs.pop(index)
                headers = resp_headers(_get(v, "headers", default={}))
                if inherit_resp.headers:
                    headers = _remove_duplicates(inherit_resp.headers, headers)
                    # if headers:
                    #     headers.extend(inherit_resp.headers)
                    # else:
                    #     headers = inherit_resp.headers
                body = resp_body(_get(v, "body", {}))
                if inherit_resp.body:
                    body = _remove_duplicates(inherit_resp.body, body)
                    # if body:
                    #     body.extend(inherit_resp.body)
                    # else:
                    #     body = inherit_resp.body
                resp = Response(
                    code=k,
                    raw={k: v},  # should prob get data union
                    method=method,
                    desc=_get(v, "description") or inherit_resp.desc,
                    headers=headers,
                    body=body,
                    config=root.config,
                    errors=root.errors
                )
                resp_objs.insert(index, resp)  # preserve order
            else:
                _headers = _get(v, "headers", default={})
                _body = _get(v, "body", default={})
                resp = Response(
                    code=k,
                    raw={k: v},
                    method=method,
                    desc=_get(v, "description"),
                    headers=resp_headers(_headers),
                    body=resp_body(_body),
                    config=root.config,
                    errors=root.errors
                )
                resp_objs.append(resp)

        return resp_objs or None

    def uri_params():
        """Set resource's URI parameters."""
        unparsed_attr = "uriParameters"
        parsed_attr = "uri_params"
        root_params = root.uri_params
        params = _create_uri_params(unparsed_attr, parsed_attr, root_params,
                                    root, type_(), is_(), method, raw_data,
                                    parent)
        declared = []
        base = base_uri_params()
        if base:
            declared = [p.name for p in base]

        return _preserve_uri_order(absolute_uri(), params, root.config,
                                   root.errors, declared)

    def base_uri_params():
        """Set resource's base URI parameters."""
        root_params = root.base_uri_params
        kw = dict(type=type_(), is_=is_(), root_params=root_params)
        params = set_params(raw_data, "base_uri_params", root, method,
                            inherit=True, **kw)
        declared = []
        uri = root.uri_params
        base = root.base_uri_params
        if uri:
            declared = [p.name for p in uri]
        if base:
            declared.extend([p.name for p in base])
        return _preserve_uri_order(root.base_uri, params, root.config,
                                   root.errors, declared)

    def query_params():
        kw = dict(type_=type_(), is_=is_())
        return set_params(raw_data, "query_params", root, method,
                          inherit=True, **kw)

    def form_params():
        """Set resource's form parameters."""
        kw = dict(type_=type_(), is_=is_())
        return set_params(raw_data, "form_params", root, method,
                          inherit=True, **kw)

    def media_type_():
        """Set resource's supported media types."""
        if method is None:
            return None
        kwargs = dict(root=root,
                      is_=is_(),
                      type_=type_(),
                      method=method,
                      data=raw_data)
        objects_to_inherit = [
            "method", "traits", "types", "resource", "root"
        ]
        inherited = get_inherited("mediaType", objects_to_inherit, **kwargs)
        meth = inherited.get("method")
        trait = inherited.get("trait")
        r_type = inherited.get("types")
        res = inherited.get("resource")
        root_ = inherited.get("root")
        return meth or trait or r_type or res or root_

    def description():
        """Set resource's description."""
        desc = _get(raw_data, "description")
        try:
            desc = _get(_get(raw_data, method), "description")
            if desc is None:
                raise AttributeError
        except AttributeError:
            if type_():
                assigned = _resource_type_lookup(type_(), root)
                try:
                    if assigned.method == method:
                        desc = assigned.description.raw
                except AttributeError:
                    pass
            else:
                desc = _get(raw_data, "description")
        return desc

    def is_():
        """Set resource's assigned trait names."""
        is_list = []
        res_level = _get(raw_data, "is")
        if res_level:
            assert isinstance(res_level, list), "Error parsing trait"
            is_list.extend(res_level)
        method_level = _get(raw_data, method, {})
        if method_level:
            method_level = _get(method_level, "is")
            if method_level:
                assert isinstance(method_level, list), "Error parsing trait"
                is_list.extend(method_level)
        return is_list or None

    def traits():
        """Set resource's assigned trait objects."""
        assigned = is_()
        if assigned:
            if root.traits:
                trait_objs = []
                for trait in assigned:
                    obj = [t for t in root.traits if t.name == trait]
                    if obj:
                        trait_objs.append(obj[0])
                return trait_objs or None

    # TODO: wow this function sucks.
    def type_():
        """Set resource's assigned resource type name."""
        __get_method = _get(raw_data, method, {})
        assigned_type = _get(__get_method, "type")
        if assigned_type:
            if not isinstance(assigned_type, dict):
                return assigned_type
            return list(iterkeys(assigned_type))[0]  # NOCOV

        assigned_type = _get(raw_data, "type")
        if isinstance(assigned_type, dict):
            return list(iterkeys(assigned_type))[0]  # NOCOV
        return assigned_type

    def resource_type():
        """Set resource's assigned resource type objects."""
        if type_() and root.resource_types:
            assigned_name = type_()
            res_types = root.resource_types
            type_obj = [r for r in res_types if r.name == assigned_name]
            if type_obj:
                return type_obj[0]

    def secured_by():
        """
        Set resource's assigned security scheme names and related paramters.
        """
        if method is not None:
            method_level = _get(raw_data, method, {})
            if method_level:
                secured_by = _get(method_level, "securedBy")
                if secured_by:
                    return secured_by
        resource_level = _get(raw_data, "securedBy")
        if resource_level:
            return resource_level
        root_level = root.secured_by
        if root_level:
            return root_level

    def security_schemes_():
        """Set resource's assigned security scheme objects."""
        secured = secured_by()
        return security_schemes(secured, root)

    node = ResourceNode(
        name=name,
        raw=raw_data,
        method=method,
        parent=parent,
        root=root,
        display_name=_get(raw_data, "displayName", name),
        path=path(),
        absolute_uri=absolute_uri(),
        protocols=protocols(),
        headers=headers(),
        body=body(),
        responses=responses(),
        uri_params=uri_params(),
        base_uri_params=base_uri_params(),
        query_params=query_params(),
        form_params=form_params(),
        media_type=media_type_(),
        desc=description(),
        is_=is_(),
        traits=traits(),
        type=type_(),
        resource_type=resource_type(),
        secured_by=secured_by(),
        security_schemes=security_schemes_(),
        errors=root.errors
    )
    if resource_type():
        # correct inheritance (issue #23)
        node._inherit_type()
    return node



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB
from __future__ import absolute_import, division, print_function

import os

import json
import pytest
from six import iteritems

from ramlfications import loader
from ramlfications.errors import LoadRAMLError

from .base import EXAMPLES, JSONREF
from .data.fixtures import load_fixtures as lf


def dict_equal(dict1, dict2):
    for k, v1 in list(iteritems(dict1)):
        assert k in dict2
        v2 = dict2[k]
        assert v1 == v2
    return True


def test_load_file():
    raml_file = os.path.join(EXAMPLES + "base-includes.raml")
    with open(raml_file) as f:
        raml = loader.RAMLLoader().load(f)

    expected_data = lf.load_file_expected_data
    assert dict_equal(raml, expected_data)


def test_load_file_with_nested_includes():
    raml_file = os.path.join(EXAMPLES + "nested-includes.raml")
    with open(raml_file) as f:
        raml = loader.RAMLLoader().load(f)

    expected_data = lf.load_file_with_nested_includes_expected
    assert dict_equal(raml, expected_data)


def test_load_file_with_nonyaml_include():
    raml_file = os.path.join(EXAMPLES + "nonyaml-includes.raml")
    with open(raml_file) as f:
        raml = loader.RAMLLoader().load(f)

    expected_data = {
        'not_yaml': "This is just a string.\n",
        'title': 'GitHub API Demo - Includes',
        'version': 'v3'
    }

    assert dict_equal(raml, expected_data)


def test_load_string():
    raml_str = ("""
                - foo
                - bar
                - baz
                """)
    raml = loader.RAMLLoader().load(raml_str)

    expected_data = ["foo", "bar", "baz"]
    assert raml.sort() == expected_data.sort()


def test_yaml_parser_error():
    raml_obj = os.path.join(EXAMPLES, "invalid_yaml.yaml")
    with pytest.raises(LoadRAMLError) as e:
        loader.RAMLLoader().load(open(raml_obj))
    msg = "Error parsing RAML:"
    assert msg in e.value.args[0]


def test_include_json():
    raml_file = os.path.join(EXAMPLES + "json_includes.raml")
    with open(raml_file) as f:
        raml = loader.RAMLLoader().load(f)

        expected_data = lf.include_json_expected
        assert dict_equal(raml, expected_data)


def test_include_xsd():
    raml_file = os.path.join(EXAMPLES + "xsd_includes.raml")
    with open(raml_file) as f:
        raml = loader.RAMLLoader().load(f)

    expected_data = lf.include_xsd_expected
    assert dict_equal(raml, expected_data)


def test_include_markdown():
    raml_file = os.path.join(EXAMPLES + "md_includes.raml")
    with open(raml_file) as f:
        raml = loader.RAMLLoader().load(f)

    expected_data = lf.include_markdown_expected
    assert dict_equal(raml, expected_data)


def test_invalid_yaml_tag():
    raml_file = os.path.join(EXAMPLES, "invalid_yaml_tag.raml")
    with pytest.raises(LoadRAMLError) as e:
        loader.RAMLLoader().load(open(raml_file))

    msg = "Error parsing RAML:"
    assert msg in e.value.args[0]


def test_includes_has_invalid_tag():
    raml_file = os.path.join(EXAMPLES, "include_has_invalid_tag.raml")
    with pytest.raises(LoadRAMLError) as e:
        loader.RAMLLoader().load(open(raml_file))

    msg = "Error parsing RAML:"
    assert msg in e.value.args[0]


def test_jsonref_relative_empty_fragment():
    raml_file = os.path.join(JSONREF, "jsonref_empty_fragment.raml")
    with open(raml_file) as f:
        raml = loader.RAMLLoader().load(f)

    expected_data = lf.jsonref_relative_empty_fragment_expected
    assert dict_equal(raml, expected_data)


def test_jsonref_relative_nonempty_fragment():
    raml_file = os.path.join(JSONREF, "jsonref_nonempty_fragment.raml")
    with open(raml_file) as f:
        raml = loader.RAMLLoader().load(f)

    expected_data = lf.jsonref_relative_nonempty_fragment_expected
    assert dict_equal(raml, expected_data)


def test_jsonref_internal_fragment_reference():
    raml_file = os.path.join(JSONREF, "jsonref_internal_fragment.raml")
    with open(raml_file) as f:
        raml = loader.RAMLLoader().load(f)

    expected_data = lf.jsonref_internal_fragment_reference_expected
    assert dict_equal(raml, expected_data)


def test_jsonref_multiref_internal_fragments():
    raml_file = os.path.join(JSONREF,
                             "jsonref_multiref_internal_fragment.raml")
    with open(raml_file) as f:
        raml = loader.RAMLLoader().load(f)

    expected_data = lf.jsonref_multiref_internal_fragments_expected
    assert dict_equal(raml, expected_data)


def test_jsonref_absolute_local_uri(tmpdir):
    # Set up a tmp RAML file with an absolute path
    json_schema_file = tmpdir.join("json_absolute_ref.json")
    data = lf.json_ref_absolute_jsondump
    json_schema_file.write(json.dumps(data))

    raml_file = tmpdir.join("json_absolute_ref.raml")
    output = lf.json_ref_absolute_ramlfile
    raml_file.write(output.format(json_file=json_schema_file.strpath))

    # Now load it
    raml = loader.RAMLLoader().load(raml_file.read())
    expected_data = lf.json_ref_absolute_expected
    assert dict_equal(raml, expected_data)


def test_jsonref_relative_local_uri():
    # RAML file includes a JSON schema.  The JSON schema includes a
    # reference to another local JSON file (same level)
    ramlfile = os.path.join(JSONREF, "jsonref_relative_local.raml")
    with open(ramlfile, "r") as f:
        loaded_raml = loader.RAMLLoader().load(f)

    schemas = loaded_raml.get("schemas")
    expected = lf.jsonref_relative_local_expected
    actual = schemas[0].get("jsonexample")

    assert dict_equal(expected, actual)


def test_jsonref_relative_local_uri_includes():
    # RAML file includes a JSON schema.  JSON schema includes a reference
    # to another local JSON file (in another directory)
    ramlfile = os.path.join(JSONREF, "jsonref_relative_local_includes.raml")
    with open(ramlfile, "r") as f:
        loaded_raml = loader.RAMLLoader().load(f)

    schemas = loaded_raml.get("schemas")

    expected = lf.jsonref_relative_local_includes_expected

    actual = schemas[0].get("jsonexample")
    assert dict_equal(expected, actual)


def test_jsonref_remote_uri(tmpdir, httpserver):
    mock_remote_json = os.path.join(JSONREF, "jsonref_mock_remote.json")
    httpserver.serve_content(open(mock_remote_json).read())

    # since we don't know what port httpserver will be on until it's
    # created, have to create the JSON file that ref's it with the url
    # variable & RAML
    jsonfile = tmpdir.join("jsonref_remote_url.json")
    data = lf.jsonref_remote_uri_jsondump
    mock_remote_url = httpserver.url + "#properties"
    data["properties"]["images"]["items"][0]["$ref"] = mock_remote_url
    jsonfile.write(json.dumps(data))

    ramlfile = tmpdir.join("jsonref_remote_url.raml")
    output = lf.jsonref_remote_uri_raml.format(json_file=jsonfile.strpath)
    ramlfile.write(output)
    readfile = ramlfile.read()
    loaded = loader.RAMLLoader().load(readfile)
    expected = lf.json_remote_uri_expected
    assert dict_equal(expected, loaded)


def test_jsonref_relative_local_file():
    # 'file:' prepends the local filename in the JSON schema
    ramlfile = os.path.join(JSONREF, "jsonref_relative_local_file.raml")
    with open(ramlfile, "r") as f:
        loaded_raml = loader.RAMLLoader().load(f)

    schemas = loaded_raml.get("schemas")
    expected = lf.jsonref_relative_local_file_expected
    actual = schemas[0].get("jsonexample")

    assert dict_equal(expected, actual)


def test_jsonref_relative_local_includes_file():
    # 'file:' prepends the local filename in the JSON schema
    filename = "jsonref_relative_local_includes_file.raml"
    ramlfile = os.path.join(JSONREF, filename)
    with open(ramlfile, "r") as f:
        loaded_raml = loader.RAMLLoader().load(f)

    schemas = loaded_raml.get("schemas")
    expected = lf.jsonref_relative_local_includes_expected
    actual = schemas[0].get("jsonexample")

    assert dict_equal(expected, actual)


def test_jsonref_absolute_local_uri_file(tmpdir):
    # Set up a tmp RAML file with an absolute path
    json_schema_file = tmpdir.join("json_absolute_ref_file.json")
    data = lf.json_ref_absolute_jsondump_file
    json_schema_file.write(json.dumps(data))

    raml_file = tmpdir.join("json_absolute_ref_file.raml")
    output = lf.json_ref_absolute_ramlfile
    raml_file.write(output.format(json_file=json_schema_file.strpath))

    # Now load it
    raml = loader.RAMLLoader().load(raml_file.read())
    expected_data = lf.json_ref_absolute_expected
    assert dict_equal(raml, expected_data)
